const assert = (value) => {
	if (!value) {
		throw new Error('An assertion failed')
	}
};

const CollabMap = class {
	constructor () {
		this.changeHandlers = [];
		this.localOps = new WeakSet();
		this.ops = new Map();
	}
	delete (key) {
		this.set(key, '');
	}
	get (key) {
		const opOrUndefined = this.ops.get(key);
		return opOrUndefined === undefined ? '' : opOrUndefined.value
	}
	getVersion (key) {
		const opOrUndefined = this.ops.get(key);
		return opOrUndefined === undefined ? 0 : opOrUndefined.version
	}
	has (key) {
		const opOrUndefined = this.ops.get(key);
		return opOrUndefined !== undefined && opOrUndefined.value !== ''
	}
	* keys () {
		for (const op of this.ops.values()) {
			if (op.value !== '') {
				yield op.key;
			}
		}
	}
	receive (op) {
		const existingOpOrUndefined = this.ops.get(op.key);
		if (
			existingOpOrUndefined === undefined
			|| existingOpOrUndefined.version < op.version
			|| (existingOpOrUndefined.version === op.version && existingOpOrUndefined.value < op.value)
		) {
			this.ops.set(op.key, op);
			const change = {
				key: op.key,
				oldValue: existingOpOrUndefined === undefined ? '' : existingOpOrUndefined.value,
				op: op,
				value: op.value,
			};
			for (const handler of this.changeHandlers) {
				handler(change);
			}
		}
	}
	set (key, value) {
		assert(typeof key === 'string');
		assert(typeof value === 'string');
		const existingOpOrUndefined = this.ops.get(key);
		const op = {
			key: key,
			value: value,
			version: existingOpOrUndefined === undefined ? 1 : existingOpOrUndefined.version + 1,
		};
		this.localOps.add(op);
		this.ops.set(key, op);
		const change = {
			key: key,
			oldValue: existingOpOrUndefined === undefined ? '' : existingOpOrUndefined.value,
			op: op,
			value: value,
		};
		for (const handler of this.changeHandlers) {
			handler(change);
		}
	}
	* [Symbol.iterator] () {
		for (const op of this.ops.values()) {
			if (op.value !== '') {
				yield [op.key, op.value];
			}
		}
	}
};

const qcsCollabMap = new CollabMap();

const QcsMediaExists = (media) => {
	return qcsCollabMap.has('media-' + media + '-info2')
};
QcsMediaExists.onChange = (fn) => {
	qcsCollabMap.changeHandlers.push((change) => {
		if (/^media-[^-]*-info2$/.test(change.key)) {
			fn();
		}
	});
};

const CollabPositionComparison = (a, b) => {
	return a < b ? -1 : (a > b ? 1 : 0)
};

const qcPlayHistoryMap = new Map();
for (const [key, value] of qcsCollabMap) {
	const matchOrNull = key.match(/^media-([^-]*)-skipBackInfo$/);
	if (matchOrNull !== null) {
		const [, track] = matchOrNull;
		qcPlayHistoryMap.set(JSON.parse(value).position, {track: track});
	}
}
for (const [key, value] of qcsCollabMap) {
	const matchOrNull = key.match(/^media-([^-]*)-playlists-([^-]*)-skipBackInfo$/);
	if (matchOrNull !== null) {
		const [, track, playlist] = matchOrNull;
		qcPlayHistoryMap.set(JSON.parse(value).position, {track: track, playlist: playlist});
	}
}
qcsCollabMap.changeHandlers.push((change) => {
	const matchOrNull = change.key.match(/^media-([^-]*)-skipBackInfo$/);
	if (matchOrNull !== null) {
		const [, track] = matchOrNull;
		if (change.oldValue !== '') {
			qcPlayHistoryMap.delete(JSON.parse(change.oldValue).position);
		}
		if (change.value !== '') {
			qcPlayHistoryMap.set(JSON.parse(change.value).position, {track: track});
		}
	}
});
qcsCollabMap.changeHandlers.push((change) => {
	const matchOrNull = change.key.match(/^media-([^-]*)-playlists-([^-]*)-skipBackInfo$/);
	if (matchOrNull !== null) {
		const [, track, playlist] = matchOrNull;
		if (change.oldValue !== '') {
			qcPlayHistoryMap.delete(JSON.parse(change.oldValue).position);
		}
		if (change.value !== '') {
			qcPlayHistoryMap.set(JSON.parse(change.value).position, {track: track, playlist: playlist});
		}
	}
});
qcPlayHistoryMap.onChange = (fn) => {
	qcsCollabMap.changeHandlers.push((change) => {
		if (
			/^media-[^-]*-skipBackInfo$/.test(change.key)
			|| /^media-[^-]*-playlists-[^-]*-skipBackInfo$/.test(change.key)
		) {
			fn();
		}
	});
};

let cache$4 = null;
const QcPlayHistoryPositions = () => {
	if (cache$4 === null) {
		cache$4 = [...qcPlayHistoryMap.keys()].filter((position) => {
			const info = qcPlayHistoryMap.get(position);
			return QcsMediaExists(info.track) && (info.playlist === undefined || QcsMediaExists(info.playlist))
		});
		cache$4.sort(CollabPositionComparison);
	}
	return cache$4
};
qcPlayHistoryMap.onChange(() => {
	cache$4 = null;
});
QcsMediaExists.onChange(() => {
	cache$4 = null;
});

let random = bytes => crypto.getRandomValues(new Uint8Array(bytes));
let customRandom = (alphabet, defaultSize, getRandom) => {
  let mask = (2 << (Math.log(alphabet.length - 1) / Math.LN2)) - 1;
  let step = -~((1.6 * mask * defaultSize) / alphabet.length);
  return (size = defaultSize) => {
    let id = '';
    while (true) {
      let bytes = getRandom(step);
      let j = step;
      while (j--) {
        id += alphabet[bytes[j] & mask] || '';
        if (id.length === size) return id
      }
    }
  }
};
let customAlphabet = (alphabet, size = 21) =>
  customRandom(alphabet, size, random);

const collabThread = customAlphabet('0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ', 13)();

const DecCollabIdInt = (collabIdInt) => {
	assert(collabIdInt !== '00000');
	let i = 4;
	while (collabIdInt[i] === '0') {
		i--;
	}
	return (
		collabIdInt.slice(0, i)
		+ (collabIdInt[i] === 'A' ? '9' : String.fromCharCode(collabIdInt.charCodeAt(i) - 1))
		+ 'Z'.repeat(4 - i)
	)
};

const IncCollabIdInt = (collabIdInt) => {
	assert(collabIdInt !== 'ZZZZZ');
	let i = 4;
	while (collabIdInt[i] === 'Z') {
		i--;
	}
	return (
		collabIdInt.slice(0, i)
		+ (collabIdInt[i] === '9' ? 'A' : String.fromCharCode(collabIdInt.charCodeAt(i) + 1))
		+ '0'.repeat(4 - i)
	)
};

let time$2 = '00000';
const Part = (int) => {
	const part = collabThread + int + time$2 + '1';
	time$2 = IncCollabIdInt(time$2);
	return part
};
const SiblingAfter = (position, i) => {
	return position.slice(0, i) + Part(IncCollabIdInt(position.slice(i + 13, i + 18)))
};
const SiblingBefore = (position, i) => {
	return position.slice(0, i) + Part(DecCollabIdInt(position.slice(i + 13, i + 18)))
};
const ChildAfter = (position) => {
	return position + Part('I0000')
};
const ChildBefore = (position) => {
	return position.slice(0, -1) + '0' + Part('I0000')
};
const SiblingOrChildAfter = (position, i) => {
	for (let j = i; j < position.length; j += 24) {
		if (position.slice(j, j + 13) === collabThread) {
			return SiblingAfter(position, j)
		}
	}
	return ChildAfter(position)
};
const SiblingOrChildBefore = (position, i) => {
	for (let j = i; j < position.length; j += 24) {
		if (position.slice(j, j + 13) === collabThread) {
			return SiblingBefore(position, j)
		}
	}
	return ChildBefore(position)
};
const CollabPosition = (afterPositionOrNull, beforePositionOrNull) => {
	if (afterPositionOrNull !== null && beforePositionOrNull !== null) {
		const minLength = Math.min(afterPositionOrNull.length, beforePositionOrNull.length);
		for (let i = 0; i < minLength; i += 24) {
			const afterThread = afterPositionOrNull.slice(i, i + 13);
			const beforeThread = beforePositionOrNull.slice(i, i + 13);
			if (afterThread !== beforeThread) {
				if (afterThread === collabThread) {
					return SiblingAfter(afterPositionOrNull, i)
				}
				if (beforeThread === collabThread) {
					return SiblingBefore(beforePositionOrNull, i)
				}
			}
		}
		if (afterPositionOrNull.length > beforePositionOrNull.length) {
			return SiblingOrChildAfter(afterPositionOrNull, beforePositionOrNull.length)
		}
		if (beforePositionOrNull.length > afterPositionOrNull.length) {
			return SiblingOrChildBefore(beforePositionOrNull, afterPositionOrNull.length)
		}
		return ChildAfter(afterPositionOrNull)
	}
	if (afterPositionOrNull !== null) {
		return SiblingOrChildAfter(afterPositionOrNull, 0)
	}
	if (beforePositionOrNull !== null) {
		return SiblingOrChildBefore(beforePositionOrNull, 0)
	}
	return Part('I0000')
};

const qcQueueMap = new Map();
for (const [key, value] of qcsCollabMap) {
	const matchOrNull = key.match(/^media-([^-]*)-queueInfo$/);
	if (matchOrNull !== null) {
		const [, track] = matchOrNull;
		const info = JSON.parse(value);
		qcQueueMap.set(info.position, {track: track, bumpNumber: info.bumpNumber === undefined ? 1 : info.bumpNumber});
	}
}
for (const [key, value] of qcsCollabMap) {
	const matchOrNull = key.match(/^media-([^-]*)-playlists-([^-]*)-queueInfo$/);
	if (matchOrNull !== null) {
		const [, track, playlist] = matchOrNull;
		const info = JSON.parse(value);
		qcQueueMap.set(info.position, {track: track, playlist: playlist, bumpNumber: info.bumpNumber === undefined ? 1 : info.bumpNumber});
	}
}
qcsCollabMap.changeHandlers.push((change) => {
	const matchOrNull = change.key.match(/^media-([^-]*)-queueInfo$/);
	if (matchOrNull !== null) {
		const [, track] = matchOrNull;
		if (change.oldValue !== '') {
			qcQueueMap.delete(JSON.parse(change.oldValue).position);
		}
		if (change.value !== '') {
			const info = JSON.parse(change.value);
			qcQueueMap.set(info.position, {track: track, bumpNumber: info.bumpNumber === undefined ? 1 : info.bumpNumber});
		}
	}
});
qcsCollabMap.changeHandlers.push((change) => {
	const matchOrNull = change.key.match(/^media-([^-]*)-playlists-([^-]*)-queueInfo$/);
	if (matchOrNull !== null) {
		const [, track, playlist] = matchOrNull;
		if (change.oldValue !== '') {
			qcQueueMap.delete(JSON.parse(change.oldValue).position);
		}
		if (change.value !== '') {
			const info = JSON.parse(change.value);
			qcQueueMap.set(info.position, {track: track, playlist: playlist, bumpNumber: info.bumpNumber === undefined ? 1 : info.bumpNumber});
		}
	}
});
qcQueueMap.onChange = (fn) => {
	qcsCollabMap.changeHandlers.push((change) => {
		if (
			/^media-[^-]*-queueInfo$/.test(change.key)
			|| /^media-[^-]*-playlists-[^-]*-queueInfo$/.test(change.key)
		) {
			fn();
		}
	});
};

const qcRecentsMap = new Map();
for (const [key, value] of qcsCollabMap) {
	const matchOrNull = key.match(/^media-([^-]*)-recentsInfo$/);
	if (matchOrNull !== null) {
		const [, media] = matchOrNull;
		qcRecentsMap.set(JSON.parse(value).position, media);
	}
}
qcsCollabMap.changeHandlers.push((change) => {
	const matchOrNull = change.key.match(/^media-([^-]*)-recentsInfo$/);
	if (matchOrNull !== null) {
		const [, media] = matchOrNull;
		if (change.oldValue !== '') {
			qcRecentsMap.delete(JSON.parse(change.oldValue).position);
		}
		if (change.value !== '') {
			qcRecentsMap.set(JSON.parse(change.value).position, media);
		}
	}
});
qcRecentsMap.onChange = (fn) => {
	qcsCollabMap.changeHandlers.push((change) => {
		if (/^media-[^-]*-recentsInfo$/.test(change.key)) {
			fn();
		}
	});
};

let cache$3 = null;
const QcRecentsPositions = () => {
	if (cache$3 === null) {
		cache$3 = [...qcRecentsMap.keys()].filter((position) => {
			return QcsMediaExists(qcRecentsMap.get(position))
		});
		cache$3.sort(CollabPositionComparison);
	}
	return cache$3
};
qcRecentsMap.onChange(() => {
	cache$3 = null;
});
QcsMediaExists.onChange(() => {
	cache$3 = null;
});

let cache$2 = null;
const QcQueuePositions = () => {
	if (cache$2 === null) {
		cache$2 = [...qcQueueMap.keys()].filter((position) => {
			const value = qcQueueMap.get(position);
			return QcsMediaExists(value.track) && (value.playlist === undefined || QcsMediaExists(value.playlist))
		});
		cache$2.sort(CollabPositionComparison);
	}
	return cache$2
};
qcQueueMap.onChange(() => {
	cache$2 = null;
});
QcsMediaExists.onChange(() => {
	cache$2 = null;
});

const qcSkipForward = () => {
	const queuePositions = QcQueuePositions();
	if (queuePositions.length) {
		const currentPlaylistOrUndefined = qcQueueMap.get(queuePositions[0]).playlist;
		const currentTrack = qcQueueMap.get(queuePositions[0]).track;
		const skipBackPositions = QcPlayHistoryPositions();
		const recentsPositions = QcRecentsPositions();
		if (currentPlaylistOrUndefined === undefined) {
			const queueInfo = JSON.parse(qcsCollabMap.get('media-' + currentTrack + '-queueInfo'));
			qcsCollabMap.delete('media-' + currentTrack + '-queueInfo');
			qcsCollabMap.set('media-' + currentTrack + '-skipBackInfo', JSON.stringify({...queueInfo, position: CollabPosition(skipBackPositions.length ? skipBackPositions[skipBackPositions.length - 1] : null, null)}));
			qcsCollabMap.set('media-' + currentTrack + '-recentsInfo', JSON.stringify({position: CollabPosition(null, recentsPositions.length === 0 ? null : recentsPositions[0]), time: Date.now()}));
		}
		else {
			const currentPlaylist = currentPlaylistOrUndefined;
			const queueInfo = JSON.parse(qcsCollabMap.get('media-' + currentTrack + '-playlists-' + currentPlaylist + '-queueInfo'));
			qcsCollabMap.delete('media-' + currentTrack + '-playlists-' + currentPlaylist + '-queueInfo');
			qcsCollabMap.set('media-' + currentTrack + '-playlists-' + currentPlaylist + '-skipBackInfo', JSON.stringify({...queueInfo, position: CollabPosition(skipBackPositions.length ? skipBackPositions[skipBackPositions.length - 1] : null, null)}));
			qcsCollabMap.set('media-' + currentPlaylist + '-recentsInfo', JSON.stringify({position: CollabPosition(null, recentsPositions.length === 0 ? null : recentsPositions[0]), time: Date.now()}));
		}
		const playCountKey = 'media-' + currentTrack + '-playCount';
		const playCount = qcsCollabMap.has(playCountKey) ? JSON.parse(qcsCollabMap.get(playCountKey)) : 0;
		qcsCollabMap.set(playCountKey, JSON.stringify(playCount + 1));
	}
};

const QcQueueVolume = () => {
	return qcsCollabMap.has('queueVolume') ? JSON.parse(qcsCollabMap.get('queueVolume')) : 1
};
QcQueueVolume.set = (volume) => {
	qcsCollabMap.set('queueVolume', JSON.stringify(volume));
};

const QcQueuePlayInfo = () => {
	return qcsCollabMap.has('queuePlayInfo') ? JSON.parse(qcsCollabMap.get('queuePlayInfo')) : {isPlaying: false}
};
QcQueuePlayInfo.set = (info) => {
	qcsCollabMap.set('queuePlayInfo', JSON.stringify(info));
};

const QcQueueIsPlaying = () => {
	return QcQueuePlayInfo().isPlaying
};
QcQueueIsPlaying.set = (isPlaying) => {
	assert(typeof isPlaying === 'boolean');
	const queuePositions = QcQueuePositions();
	QcQueuePlayInfo.set({
		isPlaying: isPlaying,
		sourceTrack: queuePositions.length ? qcQueueMap.get(queuePositions[0]).track : undefined,
		sourcePlaylist: queuePositions.length ? qcQueueMap.get(queuePositions[0]).playlist : undefined,
	});
};

const QcsMediaInfoOrNull = (media) => {
	const infoAsJsonOrEmptyString = qcsCollabMap.get('media-' + media + '-info2');
	return infoAsJsonOrEmptyString === '' ? null : JSON.parse(infoAsJsonOrEmptyString)
};

const QueryString = (initOrUndefined) => {
	return (new URLSearchParams(initOrUndefined)).toString()
};

const DateAsJsonValue = (date) => {
	return {time: date.getTime(), timezoneOffset: date.getTimezoneOffset()}
};

const scriptEl = document.getElementById('qcSetupArgScriptEl');
const qcSetupArg = JSON.parse(scriptEl.text);
scriptEl.remove();

let time$1 = 0;
const qcWriteDebuggingInfo = (kind, value) => {
	const promise = fetch('/writeDebuggingInfo', {
		body: JSON.stringify({
			collabThread: collabThread,
			kind: kind,
			performanceNow: performance.now(),
			stack: (new Error()).stack,
			time: time$1,
			value: value,
		}),
		method: 'POST',
	});
	time$1++;
	return promise
};

const ThroweeAsJsonValue = (throwee) => {
	try {
		return JSON.parse(JSON.stringify({
			code: throwee.code,
			message: throwee.message,
			name: throwee.name,
			signal: throwee.signal,
			stack: throwee.stack,
		}))
	}
	catch (anotherThrowee) {
		return null
	}
};

const postAnimators = new WeakMap();

const animators = new WeakMap();

const animateEl = (el) => {
	if (animators.has(el)) {
		animators.get(el)();
	}
	for (const child of el.children) {
		animateEl(child);
	}
	if (postAnimators.has(el)) {
		postAnimators.get(el)();
	}
};
let isAnimated = false;
const animate = () => {
	if (!isAnimated) {
		requestAnimationFrame(() => {
			isAnimated = false;
			animateEl(document.body);
		});
		isAnimated = true;
	}
};

for (const op of qcSetupArg.collabMapOps) {
	qcsCollabMap.receive(op);
}
qcsCollabMap.changeHandlers.push(animate);

const RefFn = (initialValue) => {
	let value = initialValue;
	const fn = () => {
		return value
	};
	fn.set = (newValue) => {
		value = newValue;
	};
	return fn
};

const QcMountedMainOrNull = RefFn(null);

const qcFetchOkJsonOrNull = async (resource, initOrUndefined) => {
	try {
		const response = await fetch(resource, initOrUndefined);
		return response.ok ? await response.json() : null
	}
	catch (throwee) {
		qcWriteDebuggingInfo('qcFetchOkJsonOrNullThrowee', ThroweeAsJsonValue(throwee));
		return null
	}
};

const QcButton = () => {
	const button = document.createElement('button');
	button.classList.add('button');
	return button
};

const setElChild = (el, child) => {
	if (el.firstChild !== child) {
		el.replaceChildren(child);
	}
};

const QcExistingBumpNumber = (newPosition, currentBumpNumber) => {
	const positions = QcQueuePositions();
	assert(positions.length !== 0);
	const newIndex = positions.findLastIndex((position) => {
		return newPosition > position
	}) + 1;
	if (newIndex === 0) {
		return qcQueueMap.get(positions[0]).bumpNumber
	}
	if (newIndex === positions.length) {
		return qcQueueMap.get(positions[positions.length - 1]).bumpNumber
	}
	if (currentBumpNumber === qcQueueMap.get(positions[newIndex - 1]).bumpNumber || currentBumpNumber === qcQueueMap.get(positions[newIndex]).bumpNumber) {
		return currentBumpNumber
	}
	return qcQueueMap.get(positions[newIndex - 1]).bumpNumber
};

const QcBumpNumber = (position, bumpGroups) => {
	const groupIndexOrNegativeOne = bumpGroups.findLastIndex((group) => {
		return group.positions.some((groupPosition) => {
			return position > groupPosition
		})
	});
	for (let i = groupIndexOrNegativeOne === -1 ? 0 : groupIndexOrNegativeOne; i < bumpGroups.length; i++) {
		if (bumpGroups[i].number !== undefined) {
			return bumpGroups[i].number + (i - groupIndexOrNegativeOne)
		}
	}
	return bumpGroups.length - groupIndexOrNegativeOne
};

const NumberGroups = (positions) => {
	const groups = [];
	for (const position of positions) {
		const number = qcQueueMap.get(position).bumpNumber;
		if (groups.length !== 0 && number === groups.at(-1).number) {
			groups.at(-1).positions.push(position);
		}
		else {
			groups.push({positions: [position], number: number});
		}
	}
	return groups
};
const QcBumpGroups = (arg) => {
	const positions = arg !== undefined && arg.withoutPositions !== undefined ? QcQueuePositions().filter((position) => {
		return !arg.withoutPositions.has(position)
	}) : QcQueuePositions();
	if (QcQueueIsPlaying()) {
		if (positions.length === 0) {
			return []
		}
		const groups = [{positions: [positions[0]]}];
		if (qcQueueMap.get(positions[0]).playlist === undefined) {
			groups.push(...NumberGroups(positions.slice(1)));
		}
		else {
			const currentPlaylist = qcQueueMap.get(positions[0]).playlist;
			const currentPlaylistLastIndex = positions.findLastIndex((position) => {
				return qcQueueMap.get(position).playlist === currentPlaylist
			});
			if (currentPlaylistLastIndex === 0) {
				groups.push(...NumberGroups(positions.slice(1)));
			}
			else {
				let currentPlaylistGroupFirstIndex = currentPlaylistLastIndex;
				while (currentPlaylistGroupFirstIndex > 1 && qcQueueMap.get(positions[currentPlaylistGroupFirstIndex - 1]).playlist === currentPlaylist) {
					currentPlaylistGroupFirstIndex--;
				}
				groups.push(...NumberGroups(positions.slice(1, currentPlaylistGroupFirstIndex)));
				groups.push({positions: positions.slice(currentPlaylistGroupFirstIndex, currentPlaylistLastIndex + 1)});
				groups.push(...NumberGroups(positions.slice(currentPlaylistLastIndex + 1)));
			}
		}
		return groups
	}
	return NumberGroups(positions)
};

const qcBumpTrack = (track) => {
	let skippedForward = false;
	if (QcsMediaExists(track)) {
		const queueInfoAsJsonOrEmptyString = qcsCollabMap.get('media-' + track + '-queueInfo');
		const positions = QcQueuePositions();
		if (queueInfoAsJsonOrEmptyString === '') {
			if (positions.length === 0) {
				QcQueueIsPlaying.set(false);
			}
			const newPosition = CollabPosition(positions.length === 0 ? null : positions[positions.length - 1], null);
			qcsCollabMap.set('media-' + track + '-queueInfo', JSON.stringify({
				position: newPosition,
				bumpNumber: QcBumpNumber(newPosition, QcBumpGroups()),
			}));
		}
		else {
			const currentPosition = JSON.parse(queueInfoAsJsonOrEmptyString).position;
			const currentIndex = positions.indexOf(currentPosition);
			assert(currentIndex !== -1);
			if (currentIndex === 0) {
				if (!QcQueueIsPlaying()) {
					QcQueueIsPlaying.set(true);
				}
			}
			else if (currentIndex === 1 && QcQueueIsPlaying()) {
				qcSkipForward();
				skippedForward = true;
			}
			else {
				const groups = QcBumpGroups();
				// confusingly in this context a position is not "in" a group if position === group.positions[0], instead it's considered to be "in" the previous group, this is so bumping is never a no-op from the user's perspective - it always moves it up in the queue, it doesn't just change its bump number which is hidden state
				const groupIndex = groups.findLastIndex((group) => {
					return group.positions.some((groupPosition) => {
						return currentPosition > groupPosition
					})
				});
				assert(groupIndex !== -1); // if groupIndex === -1, currentIndex === 0 and that case has already been handled
				const newPosition = groupIndex === 0 ? CollabPosition(null, positions[0]) : CollabPosition(groups[groupIndex - 1].positions.at(-1), groups[groupIndex].positions[0]);
				qcsCollabMap.set('media-' + track + '-queueInfo', JSON.stringify({
					position: newPosition,
					bumpNumber: QcBumpNumber(newPosition, QcBumpGroups({withoutPositions: new Set([currentPosition])})),
				}));
			}
		}
	}
	return {skippedForward: skippedForward}
};

const qcSkipBack = () => {
	const queuePositions = QcQueuePositions();
	const currentTrackOrNull = queuePositions.length ? qcQueueMap.get(queuePositions[0]).track : null;
	const currentPlaylistOrUndefined = queuePositions.length ? qcQueueMap.get(queuePositions[0]).playlist : undefined;
	const skipBackValues = QcPlayHistoryPositions().map((position) => {
		return qcPlayHistoryMap.get(position)
	});
	for (let i = skipBackValues.length - 1; i >= 0; i--) {
		const previousTrack = skipBackValues[i].track;
		const previousPlaylistOrUndefined = skipBackValues[i].playlist;
		if (currentTrackOrNull !== previousTrack || currentPlaylistOrUndefined !== previousPlaylistOrUndefined) {
			if (queuePositions.length === 0) {
				QcQueueIsPlaying.set(false);
			}
			const skipBackInfo = JSON.parse(qcsCollabMap.get(previousPlaylistOrUndefined === undefined ? 'media-' + previousTrack + '-skipBackInfo' : 'media-' + previousTrack + '-playlists-' + previousPlaylistOrUndefined + '-skipBackInfo'));
			qcsCollabMap.delete(previousPlaylistOrUndefined === undefined ? 'media-' + previousTrack + '-skipBackInfo' : 'media-' + previousTrack + '-playlists-' + previousPlaylistOrUndefined + '-skipBackInfo');
			qcsCollabMap.set(
				previousPlaylistOrUndefined === undefined ? 'media-' + previousTrack + '-queueInfo' : 'media-' + previousTrack + '-playlists-' + previousPlaylistOrUndefined + '-queueInfo',
				JSON.stringify({...skipBackInfo, position: CollabPosition(null, queuePositions.length === 0 ? null : queuePositions[0])}),
			);
			return true
		}
	}
	return false
};

const qcUnbumpTrack = (track) => {
	let skippedBack = false;
	if (QcsMediaExists(track)) {
		const queueInfoAsJsonOrEmptyString = qcsCollabMap.get('media-' + track + '-queueInfo');
		if (queueInfoAsJsonOrEmptyString !== '') {
			const positions = QcQueuePositions();
			const currentPosition = JSON.parse(queueInfoAsJsonOrEmptyString).position;
			const currentIndex = positions.indexOf(currentPosition);
			assert(currentIndex !== -1);
			let unbumped = false;
			if (currentIndex === 0 && QcQueueIsPlaying()) {
				const playInfo = QcQueuePlayInfo();
				if (playInfo.sourceTrack === track && playInfo.sourcePlaylist === undefined) {
					QcQueueIsPlaying.set(false);
					unbumped = true;
				}
				else {
					skippedBack = qcSkipBack();
					unbumped = skippedBack;
				}
			}
			if (!unbumped) {
				const groups = QcBumpGroups();
				const groupIndex = groups.findIndex((group) => {
					return group.positions.includes(currentPosition)
				});
				assert(groupIndex !== -1);
				if (groupIndex === groups.length - 1) {
					qcsCollabMap.delete('media-' + track + '-queueInfo');
				}
				else {
					const newPosition = CollabPosition(groups[groupIndex + 1].positions.at(-1), groupIndex === groups.length - 2 ? null : groups[groupIndex + 2].positions[0]);
					qcsCollabMap.set('media-' + track + '-queueInfo', JSON.stringify({
						position: newPosition,
						bumpNumber: QcBumpNumber(newPosition, QcBumpGroups({withoutPositions: new Set([currentPosition])})),
					}));
				}
			}
		}
	}
	return {skippedBack: skippedBack}
};

const AnimatedRefFn = (initialValue) => {
	let value = initialValue;
	const fn = () => {
		return value
	};
	fn.set = (newValue) => {
		value = newValue;
		animate();
	};
	return fn
};

const ArrayLastValue = (array) => {
	assert(array.length !== 0);
	return array[array.length - 1]
};

let cache$1 = null;
const QcQueueGroups = () => {
	if (cache$1 === null) {
		const positions = [...qcQueueMap.keys()].filter((position) => {
			const value = qcQueueMap.get(position);
			return QcsMediaExists(value.track) && (value.playlist === undefined || QcsMediaExists(value.playlist))
		});
		positions.sort(CollabPositionComparison);
		let currentPlaylistOrUndefined;
		cache$1 = [];
		for (const position of positions) {
			const value = qcQueueMap.get(position);
			if (currentPlaylistOrUndefined !== undefined && value.playlist === currentPlaylistOrUndefined) {
				cache$1[cache$1.length - 1].push(position);
			}
			else {
				cache$1.push([position]);
			}
			currentPlaylistOrUndefined = value.playlist;
		}
	}
	return cache$1
};
qcQueueMap.onChange(() => {
	cache$1 = null;
});
QcsMediaExists.onChange(() => {
	cache$1 = null;
});

const AnimatedSet = class {
	constructor () {
		this.set = new Set();
	}
	delete (value) {
		this.set.delete(value);
		animate();
	}
	has (value) {
		return this.set.has(value)
	}
	add (value) {
		this.set.add(value);
		animate();
	}
};

const QcMediaSubtitle = (media) => {
	return qcsCollabMap.get('media-' + media + '-subtitle')
};

const QcMediaTitle = (media) => {
	return qcsCollabMap.get('media-' + media + '-title')
};

const QcMainLiContextButton = () => {
	const button = QcButton();
	button.classList.add('mainLiContextButton');
	return button
};

const QcMainLi = () => {
	const li = document.createElement('li');
	li.classList.add('mainLi');
	return li
};

const QcTitleBarForm = () => {
	const form = document.createElement('form');
	form.classList.add('titleBarForm');
	form.action = '';
	return form
};

const QcTitleBar = () => {
	const titleBar = document.createElement('div');
	titleBar.classList.add('titleBar');
	return titleBar
};

const QcMainUl = () => {
	const ul = document.createElement('ul');
	ul.classList.add('mainUl');
	return ul
};

const QcMain = () => {
	const main = document.createElement('main');
	main.classList.add('main');
	return main
};

const QcTitleBarInput = () => {
	const input = document.createElement('input');
	input.classList.add('titleBarInput');
	input.type = 'search';
	input.setAttribute('autocorrect', 'off');
	return input
};

const PlaylistGroupMainLiRow = (playlist, tracksInGroup, expandedPlaylists, liIsFirst) => {
	const row = document.createElement('div');
	row.classList.add('mainLiRow');
	if (!liIsFirst) {
		row.appendChild((() => {
			const removeButton = QcButton();
			removeButton.addEventListener('click', () => {
				for (const track of tracksInGroup) {
					qcsCollabMap.delete('media-' + track + '-playlists-' + playlist + '-queueInfo');
				}
			});
			removeButton.appendChild(document.createTextNode('❌'));
			return removeButton
		})());
	}
	row.appendChild((() => {
		const expandButton = QcButton();
		expandButton.addEventListener('click', () => {
			if (expandedPlaylists.has(playlist)) {
				expandedPlaylists.delete(playlist);
			}
			else {
				expandedPlaylists.add(playlist);
			}
		});
		animators.set(expandButton, () => {
			expandButton.textContent = expandedPlaylists.has(playlist) ? '➖' : '➕';
		});
		return expandButton
	})());
	row.appendChild((() => {
		const contextButton = QcMainLiContextButton();
		{
			let bumpNumberOrZero = 0;
			for (const track of (liIsFirst ? tracksInGroup.slice(1) : tracksInGroup)) {
				let trackBumpNumberOrZero = 0;
				const queueInfoKey = 'media-' + track + '-playlists-' + playlist + '-queueInfo';
				if (qcsCollabMap.has(queueInfoKey)) {
					const info = JSON.parse(qcsCollabMap.get(queueInfoKey));
					trackBumpNumberOrZero = info.bumpNumber === undefined ? 1 : info.bumpNumber;
				}
				if (trackBumpNumberOrZero === 0) {
					bumpNumberOrZero = 0;
					break
				}
				if (bumpNumberOrZero === 0) {
					bumpNumberOrZero = trackBumpNumberOrZero;
				}
				else if (trackBumpNumberOrZero !== bumpNumberOrZero) {
					bumpNumberOrZero = 0;
					break
				}
			}
			contextButton.appendChild((() => {
				const bumpNumberSpan = document.createElement('span');
				bumpNumberSpan.classList.add('queueLiContextButtonBumpNumberSpan');
				bumpNumberSpan.textContent = '[' + bumpNumberOrZero.toString() + '] ';
				return bumpNumberSpan
			})());
		}
		contextButton.appendChild(document.createTextNode(QcMediaTitle(playlist)));
		const subtitle = QcMediaSubtitle(playlist);
		if (subtitle !== '') {
			contextButton.appendChild(document.createTextNode(' '));
			contextButton.appendChild((() => {
				const subtitleSpan = document.createElement('span');
				subtitleSpan.classList.add('mainLiSubtitleSpan');
				subtitleSpan.appendChild(document.createTextNode('(' + subtitle + ')'));
				return subtitleSpan
			})());
		}
		return contextButton
	})());
	if (!liIsFirst) {
		row.appendChild((() => {
			const moveDownButton = QcButton();
			moveDownButton.addEventListener('click', () => {
				const groups = QcQueueGroups();
				const groupIndexOrNegativeOne = groups.findIndex((group) => {
					const firstValue = qcQueueMap.get(group[0]);
					return firstValue.playlist === playlist && firstValue.track === tracksInGroup[0]
				});
				if (groupIndexOrNegativeOne !== -1) {
					const groupIndex = groupIndexOrNegativeOne;
					if (groupIndex !== groups.length - 1) {
						let afterPosition = ArrayLastValue(groups[groupIndex + 1]);
						const beforePositionOrNull = groupIndex === groups.length - 2 ? null : groups[groupIndex + 2][0];
						let bumpNumberOrNull = null;
						for (const track of groups[groupIndex].map((position) => {
							return qcQueueMap.get(position).track
						})) {
							const newPosition = CollabPosition(afterPosition, beforePositionOrNull);
							if (bumpNumberOrNull === null) {
								bumpNumberOrNull = QcExistingBumpNumber(newPosition, qcQueueMap.get(groups[groupIndex][0]).bumpNumber);
							}
							qcsCollabMap.set(
								'media-' + track + '-playlists-' + playlist + '-queueInfo',
								JSON.stringify({position: newPosition, bumpNumber: bumpNumberOrNull}),
							);
							afterPosition = newPosition;
						}
					}
				}
			});
			moveDownButton.appendChild(document.createTextNode('🔽'));
			return moveDownButton
		})());
		row.appendChild((() => {
			const moveUpButton = QcButton();
			moveUpButton.addEventListener('click', () => {
				const groups = QcQueueGroups();
				const groupIndexOrNegativeOne = groups.findIndex((group) => {
					const firstValue = qcQueueMap.get(group[0]);
					return firstValue.playlist === playlist && firstValue.track === tracksInGroup[0]
				});
				if (groupIndexOrNegativeOne !== -1) {
					const groupIndex = groupIndexOrNegativeOne;
					if (groupIndex !== 0) {
						let afterPositionOrNull;
						let beforePosition;
						if (groupIndex === 1 && groups[0].length > 1) {
							afterPositionOrNull = groups[0][0];
							beforePosition = groups[0][1];
						}
						else {
							afterPositionOrNull = groupIndex === 1 ? null : ArrayLastValue(groups[groupIndex - 2]);
							beforePosition = groups[groupIndex - 1][0];
						}
						let bumpNumberOrNull = null;
						for (const track of groups[groupIndex].map((position) => {
							return qcQueueMap.get(position).track
						})) {
							const newPosition = CollabPosition(afterPositionOrNull, beforePosition);
							if (bumpNumberOrNull === null) {
								bumpNumberOrNull = QcExistingBumpNumber(newPosition, qcQueueMap.get(groups[groupIndex][0]).bumpNumber);
							}
							qcsCollabMap.set(
								'media-' + track + '-playlists-' + playlist + '-queueInfo',
								JSON.stringify({position: newPosition, bumpNumber: bumpNumberOrNull}),
							);
							afterPositionOrNull = newPosition;
						}
					}
				}
			});
			moveUpButton.appendChild(document.createTextNode('🔼'));
			return moveUpButton
		})());
	}
	return row
};
const PlaylistGroupMainLiTracksUl = (playlist, tracksInGroup, expandedPlaylists, liIsFirst) => {
	const tracksUl = document.createElement('ul');
	tracksUl.classList.add('mainLiUl');
	animators.set(tracksUl, () => {
		tracksUl.style.display = expandedPlaylists.has(playlist) ? '' : 'none';
	});
	for (const track of (liIsFirst ? tracksInGroup.slice(1) : tracksInGroup)) {
		const queueInfoKey = 'media-' + track + '-playlists-' + playlist + '-queueInfo';
		tracksUl.appendChild((() => {
			const trackLi = QcMainLi();
			trackLi.appendChild((() => {
				const row = document.createElement('div');
				row.classList.add('mainLiRow');
				row.appendChild((() => {
					const removeButton = QcButton();
					removeButton.addEventListener('click', () => {
						qcsCollabMap.delete(queueInfoKey);
					});
					removeButton.appendChild(document.createTextNode('❌'));
					return removeButton
				})());
				row.appendChild((() => {
					const contextButton = QcMainLiContextButton();
					{
						let bumpNumberOrZero = 0;
						if (qcsCollabMap.has('media-' + track + '-playlists-' + playlist + '-queueInfo')) {
							const info = JSON.parse(qcsCollabMap.get('media-' + track + '-playlists-' + playlist + '-queueInfo'));
							bumpNumberOrZero = info.bumpNumber === undefined ? 1 : info.bumpNumber;
						}
						contextButton.appendChild((() => {
							const bumpNumberSpan = document.createElement('span');
							bumpNumberSpan.classList.add('queueLiContextButtonBumpNumberSpan');
							bumpNumberSpan.textContent = '[' + bumpNumberOrZero.toString() + '] ';
							return bumpNumberSpan
						})());
					}
					contextButton.appendChild(document.createTextNode(QcMediaTitle(track)));
					const subtitle = QcMediaSubtitle(track);
					if (subtitle !== '') {
						contextButton.appendChild(document.createTextNode(' '));
						contextButton.appendChild((() => {
							const subtitleSpan = document.createElement('span');
							subtitleSpan.classList.add('mainLiSubtitleSpan');
							subtitleSpan.appendChild(document.createTextNode('(' + subtitle + ')'));
							return subtitleSpan
						})());
					}
					return contextButton
				})());
				// row.appendChild((() => {
				// 	const moveDownButton = QcButton()
				// 	moveDownButton.addEventListener('click', () => {
				// 		const queueInfoAsJsonOrEmptyString = qcsCollabMap.get(queueInfoKey)
				// 		if (queueInfoAsJsonOrEmptyString !== '') {
				// 			const position = JSON.parse(queueInfoAsJsonOrEmptyString).position
				// 			for (const group of QcQueueGroups()) {
				// 				const indexInGroupOrNegativeOne = group.indexOf(position)
				// 				if (indexInGroupOrNegativeOne !== -1) {
				// 					const indexInGroup = indexInGroupOrNegativeOne
				// 					if (indexInGroup !== group.length - 1) {
				// 						// broken b/c null shouldnt be null necessarily
				// 						// qcsCollabMap.set(queueInfoKey, JSON.stringify({position: CollabPosition(group[indexInGroup + 1], indexInGroup === group.length - 2 ? null : group[indexInGroup + 2])}))
				// 					}
				// 					break
				// 				}
				// 			}
				// 		}
				// 	})
				// 	moveDownButton.appendChild(document.createTextNode('🔽'))
				// 	return moveDownButton
				// })())
				// row.appendChild((() => {
				// 	const moveUpButton = QcButton()
				// 	moveUpButton.addEventListener('click', () => {
				// 		const queueInfoAsJsonOrEmptyString = qcsCollabMap.get(queueInfoKey)
				// 		if (queueInfoAsJsonOrEmptyString !== '') {
				// 			const position = JSON.parse(queueInfoAsJsonOrEmptyString).position
				// 			for (const group of QcQueueGroups()) {
				// 				const indexInGroupOrNegativeOne = group.indexOf(position)
				// 				if (indexInGroupOrNegativeOne !== -1) {
				// 					const indexInGroup = indexInGroupOrNegativeOne
				// 					if (indexInGroup > (liIsFirst ? 1 : 0)) {
				// 						// broken b/c null shouldnt be null necessarily
				// 						// qcsCollabMap.set(queueInfoKey, JSON.stringify({position: CollabPosition(indexInGroup === 1 ? null : group[indexInGroup - 2], group[indexInGroup - 1])}))
				// 					}
				// 					break
				// 				}
				// 			}
				// 		}
				// 	})
				// 	moveUpButton.appendChild(document.createTextNode('🔼'))
				// 	return moveUpButton
				// })())
				return row
			})());
			return trackLi
		})());
	}
	return tracksUl
};
const QcQueuePage = (tabAfterTick) => {
	const page = {
		titleBar: (() => {
			const bar = QcTitleBar();
			bar.appendChild((() => {
				const form = QcTitleBarForm();
				form.appendChild((() => {
					const input = QcTitleBarInput();
					input.disabled = true;
					return input
				})());
				return form
			})());
			return bar
		})(),
		main: (() => {
			const main = QcMain();
			main.appendChild((() => {
				const ul = QcMainUl();
				const MaxPlayHistoryLiCount = AnimatedRefFn(0);
				const expandedPlaylists = new AnimatedSet();
				let animatedMaxPlayHistoryLiCountOrNull = null;
				let animatedPlayHistoryPositionsOrNull = null;
				let animatedGroupsOrNull = null;
				animators.set(ul, () => {
					const maxPlayHistoryLiCount = MaxPlayHistoryLiCount();
					const playHistoryPositions = QcPlayHistoryPositions();
					const groups = QcQueueGroups();
					if (animatedMaxPlayHistoryLiCountOrNull !== maxPlayHistoryLiCount || animatedPlayHistoryPositionsOrNull !== playHistoryPositions || animatedGroupsOrNull !== groups) {
						const children = [];
						// children.push((() => {
						// 	const li = QcMainLi()
						// 	li.classList.add('queuePlayHistoryNavigationLi')
						// 	li.appendChild((() => {
						// 		const row = document.createElement('div')
						// 		row.classList.add('mainLiRow')
						// 		row.classList.add('queuePlayHistoryNavigationLiRow')
						// 		row.appendChild((() => {
						// 			const navUpButton = QcButton()
						// 			navUpButton.addEventListener('click', () => {
						// 				if (MaxPlayHistoryLiCount() < QcPlayHistoryPositions().length) {
						// 					MaxPlayHistoryLiCount.set(MaxPlayHistoryLiCount() + 1)
						// 				}
						// 			})
						// 			navUpButton.textContent = '⬆️'
						// 			return navUpButton
						// 		})())
						// 		row.appendChild((() => {
						// 			const navDownButton = QcButton()
						// 			navDownButton.addEventListener('click', () => {
						// 				const liCount = Math.min(MaxPlayHistoryLiCount(), QcPlayHistoryPositions().length)
						// 				if (liCount > 0) {
						// 					MaxPlayHistoryLiCount.set(liCount - 1)
						// 				}
						// 			})
						// 			navDownButton.textContent = '⬇️'
						// 			return navDownButton
						// 		})())
						// 		return row
						// 	})())
						// 	return li
						// })())
						const playHistoryLiCount = Math.min(maxPlayHistoryLiCount, playHistoryPositions.length);
						for (let i = playHistoryPositions.length - playHistoryLiCount; i < playHistoryPositions.length; i++) {
							const track = qcPlayHistoryMap.get(playHistoryPositions[i]).track;
							const playlistOrUndefined = qcPlayHistoryMap.get(playHistoryPositions[i]).playlist;
							children.push((() => {
								const li = QcMainLi();
								li.appendChild((() => {
									const row = document.createElement('div');
									row.classList.add('mainLiRow');
									row.appendChild((() => {
										const removeButton = QcButton();
										removeButton.addEventListener('click', () => {
											qcsCollabMap.delete(playlistOrUndefined === undefined ? 'media-' + track + '-skipBackInfo' : 'media-' + track + '-playlists-' + playlistOrUndefined + '-skipBackInfo');
											const liCount = Math.min(MaxPlayHistoryLiCount(), QcPlayHistoryPositions().length);
											if (liCount > 0) {
												MaxPlayHistoryLiCount.set(liCount - 1);
											}
										});
										removeButton.appendChild(document.createTextNode('❌'));
										return removeButton
									})());
									row.appendChild((() => {
										const contextButton = QcMainLiContextButton();
										contextButton.appendChild((() => {
											const queueIndexSpan = document.createElement('span');
											queueIndexSpan.classList.add('mainLiQueueIndexSpan');
											animators.set(queueIndexSpan, () => {
												const indexOrNegativeOne = qcsCollabMap.has('media-' + track + '-queueInfo') ? QcQueuePositions().indexOf(JSON.parse(qcsCollabMap.get('media-' + track + '-queueInfo')).position) : -1;
												queueIndexSpan.textContent = indexOrNegativeOne === -1 ? '' : '[' + (QcQueueIsPlaying() ? (indexOrNegativeOne === 0 ? 'Playing' : indexOrNegativeOne.toString(10)) : (indexOrNegativeOne + 1).toString(10)) + '] ';
											});
											return queueIndexSpan
										})());
										contextButton.appendChild(document.createTextNode(QcMediaTitle(track)));
										const subtitle = QcMediaSubtitle(track);
										if (subtitle !== '') {
											contextButton.appendChild(document.createTextNode(' '));
											contextButton.appendChild((() => {
												const subtitleSpan = document.createElement('span');
												subtitleSpan.classList.add('mainLiSubtitleSpan');
												subtitleSpan.appendChild(document.createTextNode('(' + subtitle + ')'));
												return subtitleSpan
											})());
										}
										return contextButton
									})());
									row.appendChild((() => {
										const unbumpButton = QcButton();
										unbumpButton.addEventListener('click', () => {
											const info = qcUnbumpTrack(track);
											if (info.skippedBack) {
												const liCount = Math.min(MaxPlayHistoryLiCount(), QcPlayHistoryPositions().length);
												if (liCount > 0) {
													MaxPlayHistoryLiCount.set(liCount - 1);
												}
											}
										});
										animators.set(unbumpButton, () => {
											unbumpButton.style.display = qcsCollabMap.has('media-' + track + '-queueInfo') ? '' : 'none';
										});
										unbumpButton.appendChild(document.createTextNode('🔽'));
										return unbumpButton
									})());
									row.appendChild((() => {
										const bumpButton = QcButton();
										bumpButton.addEventListener('click', () => {
											const info = qcBumpTrack(track);
											if (info.skippedForward) {
												if (MaxPlayHistoryLiCount() < QcPlayHistoryPositions().length) {
													MaxPlayHistoryLiCount.set(MaxPlayHistoryLiCount() + 1);
												}
											}
										});
										bumpButton.appendChild(document.createTextNode('🔼'));
										return bumpButton
									})());
									return row
								})());
								return li
							})());
						}
						if (groups.length) {
							children.push((() => {
								const tracksInGroup = groups[0].map((position) => {
									return qcQueueMap.get(position).track
								});
								const playlistOrUndefined = qcQueueMap.get(groups[0][0]).playlist;
								const firstTrack = tracksInGroup[0];
								const li = QcMainLi();
								li.appendChild((() => {
									const row = document.createElement('div');
									row.classList.add('mainLiRow');
									row.appendChild((() => {
										const removeButton = QcButton();
										removeButton.addEventListener('click', () => {
											if (playlistOrUndefined === undefined) {
												qcsCollabMap.delete('media-' + firstTrack + '-queueInfo');
											}
											else {
												for (const track of tracksInGroup) {
													qcsCollabMap.delete('media-' + track + '-playlists-' + playlistOrUndefined + '-queueInfo');
												}
											}
										});
										removeButton.appendChild(document.createTextNode('❌'));
										return removeButton
									})());
									row.appendChild((() => {
										const contextButton = QcMainLiContextButton();
										contextButton.classList.add('queueCurrentLiContextButton');
										{
											let bumpNumberOrZero = 0;
											const queueInfoKey = playlistOrUndefined === undefined ? 'media-' + firstTrack + '-queueInfo' : 'media-' + firstTrack + '-playlists-' + playlistOrUndefined + '-queueInfo';
											if (qcsCollabMap.has(queueInfoKey)) {
												const info = JSON.parse(qcsCollabMap.get(queueInfoKey));
												bumpNumberOrZero = info.bumpNumber === undefined ? 1 : info.bumpNumber;
											}
											contextButton.appendChild((() => {
												const bumpNumberSpan = document.createElement('span');
												bumpNumberSpan.classList.add('queueLiContextButtonBumpNumberSpan');
												bumpNumberSpan.textContent = '[' + bumpNumberOrZero.toString() + '] ';
												return bumpNumberSpan
											})());
										}
										contextButton.appendChild(document.createTextNode(QcMediaTitle(firstTrack)));
										const subtitle = QcMediaSubtitle(firstTrack);
										if (subtitle !== '') {
											contextButton.appendChild(document.createTextNode(' '));
											contextButton.appendChild((() => {
												const subtitleSpan = document.createElement('span');
												subtitleSpan.classList.add('mainLiSubtitleSpan');
												subtitleSpan.appendChild(document.createTextNode('(' + subtitle + ')'));
												return subtitleSpan
											})());
										}
										return contextButton
									})());
									row.appendChild((() => {
										const moveDownButton = QcButton();
										moveDownButton.addEventListener('click', () => {
											const groups = QcQueueGroups();
											const groupIndexOrNegativeOne = groups.findIndex((group) => {
												const firstValue = qcQueueMap.get(group[0]);
												return firstValue.playlist === playlistOrUndefined && firstValue.track === firstTrack
											});
											if (groupIndexOrNegativeOne !== -1) {
												const groupIndex = groupIndexOrNegativeOne;
												if (groupIndex !== groups.length - 1) {
													let afterPosition = ArrayLastValue(groups[groupIndex + 1]);
													const beforePositionOrNull = groupIndex === groups.length - 2 ? null : groups[groupIndex + 2][0];
													if (playlistOrUndefined === undefined) {
														const newPosition = CollabPosition(afterPosition, beforePositionOrNull);
														qcsCollabMap.set(
															'media-' + firstTrack + '-queueInfo',
															JSON.stringify({position: newPosition, bumpNumber: QcExistingBumpNumber(newPosition, qcQueueMap.get(groups[groupIndex][0]).bumpNumber)}),
														);
													}
													else {
														const playlist = playlistOrUndefined;
														let bumpNumberOrNull = null;
														for (const track of groups[groupIndex].map((position) => {
															return qcQueueMap.get(position).track
														})) {
															const newPosition = CollabPosition(afterPosition, beforePositionOrNull);
															if (bumpNumberOrNull === null) {
																bumpNumberOrNull = QcExistingBumpNumber(newPosition, qcQueueMap.get(groups[groupIndex][0]).bumpNumber);
															}
															qcsCollabMap.set(
																'media-' + track + '-playlists-' + playlist + '-queueInfo',
																JSON.stringify({position: newPosition, bumpNumber: bumpNumberOrNull}),
															);
															afterPosition = newPosition;
														}
													}
												}
											}
										});
										moveDownButton.appendChild(document.createTextNode('🔽'));
										return moveDownButton
									})());
									row.appendChild((() => {
										const moveUpButton = QcButton();
										moveUpButton.appendChild(document.createTextNode('🔼'));
										return moveUpButton
									})());
									return row
								})());
								li.appendChild((() => {
									const row = document.createElement('div');
									row.classList.add('mainLiRow');
									row.classList.add('queueCurrentLiSkipRow');
									row.appendChild((() => {
										const skipBackButton = QcButton();
										skipBackButton.classList.add('queueCurrentLiSkipBackButton');
										skipBackButton.addEventListener('click', () => {
											qcSkipBack();
										});
										skipBackButton.appendChild(document.createTextNode('⏮️'));
										return skipBackButton
									})());
									row.appendChild((() => {
										const playButton = QcButton();
										playButton.classList.add('queueCurrentLiPlayButton');
										playButton.addEventListener('click', () => {
											QcQueueIsPlaying.set(!QcQueueIsPlaying());
										});
										animators.set(playButton, () => {
											playButton.textContent = QcQueueIsPlaying() ? '⏸️' : '▶️';
										});
										return playButton
									})());
									row.appendChild((() => {
										const skipForwardButton = QcButton();
										skipForwardButton.classList.add('queueCurrentLiSkipForwardButton');
										skipForwardButton.addEventListener('click', () => {
											qcSkipForward();
										});
										skipForwardButton.appendChild(document.createTextNode('⏭️'));
										return skipForwardButton
									})());
									return row
								})());
								li.appendChild((() => {
									const row = document.createElement('div');
									row.classList.add('mainLiRow');
									row.classList.add('queueCurrentLiVolumeSliderRow');
									row.appendChild((() => {
										const volumeMutedIcon = QcButton();
										volumeMutedIcon.textContent = '🔇';
										return volumeMutedIcon
									})());
									row.appendChild((() => {
										const volumeSlider = document.createElement('input');
										volumeSlider.classList.add('queueCurrentLiVolumeSlider');
										volumeSlider.type = 'range';
										volumeSlider.min = 0;
										volumeSlider.max = 1;
										volumeSlider.step = 'any';
										volumeSlider.addEventListener('input', () => {
											QcQueueVolume.set(JSON.parse(volumeSlider.value));
										});
										animators.set(volumeSlider, () => {
											volumeSlider.value = QcQueueVolume().toString();
										});
										return volumeSlider
									})());
									row.appendChild((() => {
										const volumeMaxIcon = QcButton();
										volumeMaxIcon.textContent = '🔊';
										return volumeMaxIcon
									})());
									return row
								})());
								if (groups[0].length > 1) {
									assert(playlistOrUndefined !== undefined);
									li.appendChild(PlaylistGroupMainLiRow(playlistOrUndefined, tracksInGroup, expandedPlaylists, true));
									li.appendChild(PlaylistGroupMainLiTracksUl(playlistOrUndefined, tracksInGroup, expandedPlaylists, true));
								}
								return li
							})());
							for (let i = 1; i < groups.length; i++) {
								const playlistOrUndefined = qcQueueMap.get(groups[i][0]).playlist;
								if (playlistOrUndefined === undefined) {
									assert(groups[i].length === 1);
									const track = qcQueueMap.get(groups[i][0]).track;
									const queueInfoKey = 'media-' + track + '-queueInfo';
									children.push((() => {
										const li = QcMainLi();
										li.appendChild((() => {
											const row = document.createElement('div');
											row.classList.add('mainLiRow');
											row.appendChild((() => {
												const removeButton = QcButton();
												removeButton.addEventListener('click', () => {
													qcsCollabMap.delete(queueInfoKey);
												});
												removeButton.appendChild(document.createTextNode('❌'));
												return removeButton
											})());
											row.appendChild((() => {
												const contextButton = QcMainLiContextButton();
												{
													let bumpNumberOrZero = 0;
													if (qcsCollabMap.has(queueInfoKey)) {
														const info = JSON.parse(qcsCollabMap.get(queueInfoKey));
														bumpNumberOrZero = info.bumpNumber === undefined ? 1 : info.bumpNumber;
													}
													contextButton.appendChild((() => {
														const bumpNumberSpan = document.createElement('span');
														bumpNumberSpan.classList.add('queueLiContextButtonBumpNumberSpan');
														bumpNumberSpan.textContent = '[' + bumpNumberOrZero.toString() + '] ';
														return bumpNumberSpan
													})());
												}
												contextButton.appendChild(document.createTextNode(QcMediaTitle(track)));
												const subtitle = QcMediaSubtitle(track);
												if (subtitle !== '') {
													contextButton.appendChild(document.createTextNode(' '));
													contextButton.appendChild((() => {
														const subtitleSpan = document.createElement('span');
														subtitleSpan.classList.add('mainLiSubtitleSpan');
														subtitleSpan.appendChild(document.createTextNode('(' + subtitle + ')'));
														return subtitleSpan
													})());
												}
												return contextButton
											})());
											row.appendChild((() => {
												const moveDownButton = QcButton();
												moveDownButton.addEventListener('click', () => {
													const groups = QcQueueGroups();
													const groupIndexOrNegativeOne = groups.findIndex((group) => {
														const firstValue = qcQueueMap.get(group[0]);
														return firstValue.playlist === undefined && firstValue.track === track
													});
													if (groupIndexOrNegativeOne !== -1) {
														const groupIndex = groupIndexOrNegativeOne;
														if (groupIndex !== groups.length - 1) {
															const afterPosition = ArrayLastValue(groups[groupIndex + 1]);
															const beforePositionOrNull = groupIndex === groups.length - 2 ? null : groups[groupIndex + 2][0];
															const newPosition = CollabPosition(afterPosition, beforePositionOrNull);
															qcsCollabMap.set(
																'media-' + track + '-queueInfo',
																JSON.stringify({position: newPosition, bumpNumber: QcExistingBumpNumber(newPosition, qcQueueMap.get(groups[groupIndex][0]).bumpNumber)}),
															);
														}
													}
												});
												moveDownButton.appendChild(document.createTextNode('🔽'));
												return moveDownButton
											})());
											row.appendChild((() => {
												const moveUpButton = QcButton();
												moveUpButton.addEventListener('click', () => {
													const groups = QcQueueGroups();
													const groupIndexOrNegativeOne = groups.findIndex((group) => {
														const firstValue = qcQueueMap.get(group[0]);
														return firstValue.playlist === undefined && firstValue.track === track
													});
													if (groupIndexOrNegativeOne !== -1) {
														const groupIndex = groupIndexOrNegativeOne;
														if (groupIndex !== 0) {
															let afterPositionOrNull;
															let beforePosition;
															if (groupIndex === 1 && groups[0].length > 1) {
																afterPositionOrNull = groups[0][0];
																beforePosition = groups[0][1];
															}
															else {
																afterPositionOrNull = groupIndex === 1 ? null : ArrayLastValue(groups[groupIndex - 2]);
																beforePosition = groups[groupIndex - 1][0];
															}
															const newPosition = CollabPosition(afterPositionOrNull, beforePosition);
															qcsCollabMap.set(
																'media-' + track + '-queueInfo',
																JSON.stringify({position: newPosition, bumpNumber: QcExistingBumpNumber(newPosition, qcQueueMap.get(groups[groupIndex][0]).bumpNumber)}),
															);
														}
													}
												});
												moveUpButton.appendChild(document.createTextNode('🔼'));
												return moveUpButton
											})());
											return row
										})());
										return li
									})());

								}
								else {
									const playlist = playlistOrUndefined;
									const tracksInGroup = groups[i].map((position) => {
										return qcQueueMap.get(position).track
									});
									children.push((() => {
										const li = QcMainLi();
										li.appendChild(PlaylistGroupMainLiRow(playlist, tracksInGroup, expandedPlaylists, false));
										li.appendChild(PlaylistGroupMainLiTracksUl(playlist, tracksInGroup, expandedPlaylists, false));
										return li
									})());
								}
							}
						}
						else {
							children.push((() => {
								const li = QcMainLi();
								li.appendChild((() => {
									const row = document.createElement('div');
									row.classList.add('mainLiRow');
									row.appendChild((() => {
										const contextButton = QcMainLiContextButton();
										contextButton.classList.add('queueCurrentLiContextButton');
										contextButton.appendChild(document.createTextNode('Queue is empty'));
										return contextButton
									})());
									return row
								})());
								li.appendChild((() => {
									const row = document.createElement('div');
									row.classList.add('mainLiRow');
									row.classList.add('queueCurrentLiSkipRow');
									row.appendChild((() => {
										const skipBackButton = QcButton();
										skipBackButton.classList.add('queueCurrentLiSkipBackButton');
										skipBackButton.addEventListener('click', () => {
											qcSkipBack();
										});
										skipBackButton.appendChild(document.createTextNode('⏮️'));
										return skipBackButton
									})());
									row.appendChild((() => {
										const playButton = QcButton();
										playButton.classList.add('queueCurrentLiPlayButton');
										playButton.textContent = '▶️';
										return playButton
									})());
									row.appendChild((() => {
										const skipForwardButton = QcButton();
										skipForwardButton.classList.add('queueCurrentLiSkipForwardButton');
										skipForwardButton.appendChild(document.createTextNode('⏭️'));
										return skipForwardButton
									})());
									return row
								})());
								li.appendChild((() => {
									const row = document.createElement('div');
									row.classList.add('mainLiRow');
									row.classList.add('queueCurrentLiVolumeSliderRow');
									row.appendChild((() => {
										const volumeMutedIcon = QcButton();
										volumeMutedIcon.textContent = '🔇';
										return volumeMutedIcon
									})());
									row.appendChild((() => {
										const volumeSlider = document.createElement('input');
										volumeSlider.classList.add('queueCurrentLiVolumeSlider');
										volumeSlider.type = 'range';
										volumeSlider.min = 0;
										volumeSlider.max = 1;
										volumeSlider.step = 'any';
										volumeSlider.addEventListener('input', () => {
											QcQueueVolume.set(JSON.parse(volumeSlider.value));
										});
										animators.set(volumeSlider, () => {
											volumeSlider.value = QcQueueVolume().toString();
										});
										return volumeSlider
									})());
									row.appendChild((() => {
										const volumeMaxIcon = QcButton();
										volumeMaxIcon.textContent = '🔊';
										return volumeMaxIcon
									})());
									return row
								})());
								return li
							})());
						}
						ul.replaceChildren(...children);
						animatedMaxPlayHistoryLiCountOrNull = maxPlayHistoryLiCount;
						animatedPlayHistoryPositionsOrNull = playHistoryPositions;
						animatedGroupsOrNull = groups;
					}
				});
				return ul
			})());
			return () => {
				return main
			}
		})(),
	};
	return page
};

const AnimatedRef = class {
	constructor (value) {
		this.valueData = value;
	}
	get value () {
		return this.valueData
	}
	set value (value) {
		this.valueData = value;
		animate();
	}
};

const qcQueueTab = {pageRef: null};
qcQueueTab.pageRef = new AnimatedRef(QcQueuePage());

// import qcYtTab from './qcYtTab.js'

const qcCurrentTabRef = new AnimatedRef(qcQueueTab);

const skipBackPlaylist = (queuePositions, playlist) => {
	const skipBackValues = QcPlayHistoryPositions().map((position) => {
		return qcPlayHistoryMap.get(position)
	});
	for (let i = skipBackValues.length - 1; i >= 0; i--) {
		const previousPlaylistOrUndefined = skipBackValues[i].playlist;
		if (playlist !== previousPlaylistOrUndefined) {
			const previousTrack = skipBackValues[i].track;
			const skipBackInfo = JSON.parse(qcsCollabMap.get(previousPlaylistOrUndefined === undefined ? 'media-' + previousTrack + '-skipBackInfo' : 'media-' + previousTrack + '-playlists-' + previousPlaylistOrUndefined + '-skipBackInfo'));
			assert(queuePositions.length !== 0);
			qcsCollabMap.delete(previousPlaylistOrUndefined === undefined ? 'media-' + previousTrack + '-skipBackInfo' : 'media-' + previousTrack + '-playlists-' + previousPlaylistOrUndefined + '-skipBackInfo');
			qcsCollabMap.set(
				previousPlaylistOrUndefined === undefined ? 'media-' + previousTrack + '-queueInfo' : 'media-' + previousTrack + '-playlists-' + previousPlaylistOrUndefined + '-queueInfo',
				JSON.stringify({...skipBackInfo, position: CollabPosition(null, queuePositions[0])}),
			);
			return true
		}
	}
	return false
};
const qcUnbumpPlaylist = (playlist) => {
	const positions = QcQueuePositions();
	const firstIndexOrNegativeOne = positions.findIndex((position) => {
		return qcQueueMap.get(position).playlist === playlist
	});
	if (firstIndexOrNegativeOne !== -1) {
		const firstIndex = firstIndexOrNegativeOne;
		let unbumped = false;
		if (firstIndex === 0 && QcQueueIsPlaying()) {
			const playInfo = QcQueuePlayInfo();
			if (playInfo.sourcePlaylist === playlist) {
				QcQueueIsPlaying.set(false);
				unbumped = true;
			}
			else {
				unbumped = skipBackPlaylist(positions, playlist);
			}
		}
		if (!unbumped) {
			const groups = QcBumpGroups();
			const firstGroupIndex = groups.findIndex((group) => {
				return group.positions.includes(positions[firstIndex])
			});
			assert(firstGroupIndex !== -1);
			const tracksInPlaylist = positions.filter((position) => {
				return qcQueueMap.get(position).playlist === playlist
			}).map((position) => {
				return qcQueueMap.get(position).track
			});
			if (firstGroupIndex === groups.length - 1) {
				for (const track of tracksInPlaylist) {
					qcsCollabMap.delete('media-' + track + '-playlists-' + playlist + '-queueInfo');
				}
			}
			else {
				let afterPosition = groups[firstGroupIndex + 1].positions.at(-1);
				const beforePositionOrNull = firstGroupIndex === groups.length - 2 ? null : groups[firstGroupIndex + 2].positions[0];
				let bumpNumberOrNull = null;
				for (const track of tracksInPlaylist) {
					const newPosition = CollabPosition(afterPosition, beforePositionOrNull);
					if (bumpNumberOrNull === null) {
						bumpNumberOrNull = QcBumpNumber(newPosition, QcBumpGroups({withoutPositions: new Set(positions.filter((position) => {
							return qcQueueMap.get(position).playlist === playlist
						}))}));
					}
					qcsCollabMap.set(
						'media-' + track + '-playlists-' + playlist + '-queueInfo',
						JSON.stringify({position: newPosition, bumpNumber: bumpNumberOrNull}),
					);
					afterPosition = newPosition;
				}
			}
		}
	}
};

const maps = new Map();
for (const [key, value] of qcsCollabMap) {
	const matchOrNull = key.match(/^media-([^-]*)-playlists-([^-]*)-info$/);
	if (matchOrNull !== null) {
		const [, track, playlist] = matchOrNull;
		if (!maps.has(playlist)) {
			maps.set(playlist, new Map());
		}
		maps.get(playlist).set(JSON.parse(value).position, track);
	}
}
qcsCollabMap.changeHandlers.push((change) => {
	const matchOrNull = change.key.match(/^media-([^-]*)-playlists-([^-]*)-info$/);
	if (matchOrNull !== null) {
		const [, track, playlist] = matchOrNull;
		if (change.oldValue !== '') {
			maps.get(playlist).delete(JSON.parse(change.oldValue).position);
			if (maps.get(playlist).size === 0) {
				maps.delete(playlist);
			}
		}
		if (change.value !== '') {
			if (!maps.has(playlist)) {
				maps.set(playlist, new Map());
			}
			maps.get(playlist).set(JSON.parse(change.value).position, track);
		}
	}
});
const emptyMap = new Map();
const QcPlaylistMap = (playlist) => {
	return maps.has(playlist) ? maps.get(playlist) : emptyMap
};
QcPlaylistMap.onChange = (fn) => {
	qcsCollabMap.changeHandlers.push((change) => {
		if (/^media-[^-]*-playlists-[^-]*-info$/.test(change.key)) {
			fn();
		}
	});
};

const cache = new Map();
const QcPlaylistPositions = (playlist) => {
	if (!cache.has(playlist)) {
		const map = QcPlaylistMap(playlist);
		const positions = [...map.keys()].filter((position) => {
			return QcsMediaExists(map.get(position))
		});
		positions.sort(CollabPositionComparison);
		cache.set(playlist, positions);
	}
	return cache.get(playlist)
};
QcPlaylistMap.onChange(() => {
	cache.clear();
});
QcsMediaExists.onChange(() => {
	cache.clear();
});

const qcBumpPlaylist = (playlist) => {
	if (QcsMediaExists(playlist)) {
		const positions = QcQueuePositions();
		const firstIndexOrNegativeOne = positions.findIndex((position) => {
			return qcQueueMap.get(position).playlist === playlist
		});
		if (firstIndexOrNegativeOne === -1) {
			const playlistPositions = QcPlaylistPositions(playlist);
			if (playlistPositions.length !== 0) {
				if (positions.length === 0) {
					QcQueueIsPlaying.set(false);
				}
				const playlistMap = QcPlaylistMap(playlist);
				let afterPositionOrNull = positions.length === 0 ? null : positions[positions.length - 1];
				let bumpNumberOrNull = null;
				for (const playlistPosition of playlistPositions) {
					const position = CollabPosition(afterPositionOrNull, null);
					if (bumpNumberOrNull === null) {
						bumpNumberOrNull = QcBumpNumber(position, QcBumpGroups());
					}
					qcsCollabMap.set(
						'media-' + playlistMap.get(playlistPosition) + '-playlists-' + playlist + '-queueInfo',
						JSON.stringify({position: position, bumpNumber: bumpNumberOrNull}),
					);
					afterPositionOrNull = position;
				}
			}
		}
		else {
			const firstIndex = firstIndexOrNegativeOne;
			if (firstIndex === 0) {
				if (!QcQueueIsPlaying()) {
					QcQueueIsPlaying.set(true);
				}
			}
			else if (firstIndex === 1 && QcQueueIsPlaying()) {
				qcSkipForward();
			}
			else {
				const groups = QcBumpGroups();
				const firstGroupIndex = groups.findLastIndex((group) => {
					return group.positions.some((groupPosition) => {
						return positions[firstIndex] > groupPosition
					})
				});
				assert(firstGroupIndex !== -1); // if firstGroupIndex === -1, firstIndex === 0 and that case has already been handled
				let afterPositionOrNull = firstGroupIndex === 0 ? null : groups[firstGroupIndex - 1].positions.at(-1);
				const beforePosition = groups[firstGroupIndex].positions[0];
				let bumpNumberOrNull = null;
				for (const value of positions.map((position) => {
					return qcQueueMap.get(position)
				})) {
					if (value.playlist === playlist) {
						const newPosition = CollabPosition(afterPositionOrNull, beforePosition);
						if (bumpNumberOrNull === null) {
							bumpNumberOrNull = QcBumpNumber(newPosition, QcBumpGroups({withoutPositions: new Set(positions.filter((position) => {
								return qcQueueMap.get(position).playlist === playlist
							}))}));
						}
						qcsCollabMap.set(
							'media-' + value.track + '-playlists-' + playlist + '-queueInfo',
							JSON.stringify({position: newPosition, bumpNumber: bumpNumberOrNull}),
						);
						afterPositionOrNull = newPosition;
					}
				}
			}
		}
	}
};

const QcMediaExistsAndIsTrack = (media) => {
	const infoOrNull = QcsMediaInfoOrNull(media);
	return infoOrNull !== null && infoOrNull.ytVideo !== undefined
};

const QcMediaExistsAndIsPlaylist = (media) => {
	const infoOrNull = QcsMediaInfoOrNull(media);
	return infoOrNull !== null && infoOrNull.ytVideo === undefined
};

const QcTitleBarButton = () => {
	const button = QcButton();
	button.classList.add('titleBarButton');
	return button
};

const MediaMainLi = (media, playlistOrNull, expandedPlaylists) => {
	const li = QcMainLi();
	li.appendChild((() => {
		const row = document.createElement('div');
		row.classList.add('mainLiRow');
		row.appendChild((() => {
			const removeButton = QcButton();
			removeButton.addEventListener('click', () => {
				if (playlistOrNull === null) {
					qcsCollabMap.delete('media-' + media + '-recentsInfo');
				}
				else {
					qcsCollabMap.delete('media-' + media + '-playlists-' + playlistOrNull + '-info');
				}
			});
			removeButton.appendChild(document.createTextNode('❌'));
			return removeButton
		})());
		if (QcMediaExistsAndIsPlaylist(media)) {
			const playlist = media;
			row.appendChild((() => {
				const expandButton = QcButton();
				expandButton.addEventListener('click', () => {
					if (expandedPlaylists.has(playlist)) {
						expandedPlaylists.delete(playlist);
					}
					else {
						expandedPlaylists.add(playlist);
					}
				});
				animators.set(expandButton, () => {
					expandButton.textContent = expandedPlaylists.has(playlist) ? '➖' : '➕';
				});
				return expandButton
			})());
		}
		row.appendChild((() => {
			const contextButton = QcMainLiContextButton();
			if (QcMediaExistsAndIsTrack(media)) {
				const track = media;
				contextButton.appendChild((() => {
					const queueIndexSpan = document.createElement('span');
					queueIndexSpan.classList.add('mainLiQueueIndexSpan');
					animators.set(queueIndexSpan, () => {
						const indexOrNegativeOne = qcsCollabMap.has('media-' + track + '-queueInfo') ? QcQueuePositions().indexOf(JSON.parse(qcsCollabMap.get('media-' + track + '-queueInfo')).position) : -1;
						queueIndexSpan.textContent = indexOrNegativeOne === -1 ? '' : '[' + (QcQueueIsPlaying() ? (indexOrNegativeOne === 0 ? 'Playing' : indexOrNegativeOne.toString(10)) : (indexOrNegativeOne + 1).toString(10)) + '] ';
					});
					return queueIndexSpan
				})());
			}
			else if (QcMediaExistsAndIsPlaylist(media)) {
				const playlist = media;
				contextButton.appendChild((() => {
					const queueIndexSpan = document.createElement('span');
					queueIndexSpan.classList.add('mainLiQueueIndexSpan');
					animators.set(queueIndexSpan, () => {
						const queuePositions = QcQueuePositions();
						const firstIndexOrNegativeOne = queuePositions.findIndex((position) => {
							return qcQueueMap.get(position).playlist === playlist
						});
						let text = '';
						if (firstIndexOrNegativeOne !== -1) {
							const firstIndex = firstIndexOrNegativeOne;
							if (firstIndex === 0 && QcQueueIsPlaying()) {
								text = '[Playing] ';
							}
							else {
								let lastIndex;
								for (let i = queuePositions.length - 1; i >= 0; i--) {
									if (qcQueueMap.get(queuePositions[i]).playlist === playlist) {
										lastIndex = i;
										break
									}
								}
								const AdjustedIndex = (index) => {
									return QcQueueIsPlaying() ? index : index + 1
								};
								text = '[' + AdjustedIndex(firstIndex) + (firstIndex === lastIndex ? '' : '-' + AdjustedIndex(lastIndex)) + '] ';
							}
						}
						queueIndexSpan.textContent = text;
					});
					return queueIndexSpan
				})());
			}
			contextButton.appendChild(document.createTextNode(QcMediaTitle(media)));
			const subtitle = QcMediaSubtitle(media);
			if (subtitle !== '') {
				contextButton.appendChild(document.createTextNode(' '));
				contextButton.appendChild((() => {
					const subtitleSpan = document.createElement('span');
					subtitleSpan.classList.add('mainLiSubtitleSpan');
					subtitleSpan.appendChild(document.createTextNode('(' + subtitle + ')'));
					return subtitleSpan
				})());
			}
			return contextButton
		})());
		if (QcMediaExistsAndIsTrack(media)) {
			const track = media;
			row.appendChild((() => {
				const unbumpButton = QcButton();
				unbumpButton.addEventListener('click', () => {
					qcUnbumpTrack(track);
				});
				animators.set(unbumpButton, () => {
					unbumpButton.style.display = qcsCollabMap.has('media-' + track + '-queueInfo') ? '' : 'none';
				});
				unbumpButton.appendChild(document.createTextNode('🔽'));
				return unbumpButton
			})());
		}
		else if (QcMediaExistsAndIsPlaylist(media)) {
			const playlist = media;
			row.appendChild((() => {
				const unbumpButton = QcButton();
				unbumpButton.addEventListener('click', () => {
					qcUnbumpPlaylist(playlist);
				});
				animators.set(unbumpButton, () => {
					unbumpButton.style.display = QcQueuePositions().some((position) => {
						return qcQueueMap.get(position).playlist === playlist
					}) ? '' : 'none';
				});
				unbumpButton.appendChild(document.createTextNode('🔽'));
				return unbumpButton
			})());
		}
		row.appendChild((() => {
			const bumpButton = QcButton();
			bumpButton.addEventListener('click', () => {
				if (QcMediaExistsAndIsTrack(media)) {
					qcBumpTrack(media);
				}
				else if (QcMediaExistsAndIsPlaylist(media)) {
					qcBumpPlaylist(media);
				}
			});
			bumpButton.appendChild(document.createTextNode('🔼'));
			return bumpButton
		})());
		return row
	})());
	if (QcMediaExistsAndIsPlaylist(media)) {
		const playlist = media;
		li.appendChild((() => {
			const tracksUl = document.createElement('ul');
			tracksUl.classList.add('mainLiUl');
			let mountedPositionsOrNull = null;
			const innerExpandedPlaylists = new AnimatedSet();
			animators.set(tracksUl, () => {
				tracksUl.style.display = expandedPlaylists.has(playlist) ? '' : 'none';
				const positions = QcPlaylistPositions(playlist);
				if (mountedPositionsOrNull !== positions) {
					const map = QcPlaylistMap(playlist);
					tracksUl.replaceChildren(...positions.map((position) => {
						return MediaMainLi(map.get(position), playlist, innerExpandedPlaylists)
					}));
					mountedPositionsOrNull = positions;
				}
			});
			return tracksUl
		})());
	}
	return li
};
const Main = (searchQueryOrNull) => {
	const main = QcMain();
	const ul = QcMainUl();
	const emptyMessageNode = document.createTextNode('empty recents message');
	const noSearchResultsMessageNode = document.createTextNode('no search results');
	const expandedPlaylists = new AnimatedSet();
	let mountedGlobalPositionsOrNull = null;
	let mountedLocalMediaOrNull = null;
	animators.set(main, () => {
		const globalPositions = QcRecentsPositions();
		const localMedia = mountedLocalMediaOrNull === null || QcMountedMainOrNull() !== main ? globalPositions.map((position) => {
			return qcRecentsMap.get(position)
		}) : mountedLocalMediaOrNull;
		// todo this doesn't fix deleted-then-reappears-on-change-from-other-element but that's probably fine for now
		// fix this ^ (probably using something like a removedLocalMedia set) when you do a deeper refactor of this animation
		// logic which is pretty gross rn lol
		if (mountedGlobalPositionsOrNull !== globalPositions || mountedLocalMediaOrNull !== localMedia) {
			if (searchQueryOrNull === null) {
				const ulChildren = [];
				let currentTimeOrNull = null;
				let currentTimeUl;
				for (const media of localMedia.slice(0, 100)) {
					if (qcsCollabMap.has('media-' + media + '-recentsInfo') && QcsMediaExists(media)) {
						const date = new Date(JSON.parse(qcsCollabMap.get('media-' + media + '-recentsInfo')).time);
						date.setHours(0, 0, 0, 0);
						const time = date.getTime();
						if (currentTimeOrNull !== time) {
							currentTimeOrNull = time;
							currentTimeUl = QcMainUl();
							ulChildren.push((() => {
								const li = QcMainLi();
								li.appendChild((() => {
									const row = document.createElement('div');
									row.classList.add('mainLiRow');
									row.appendChild((() => {
										const contextButton = QcMainLiContextButton();
										contextButton.classList.add('recentsDateLiContextButton');
										contextButton.textContent = date.toString().slice(4, 10);
										return contextButton
									})());
									row.style.display = 'none';// turning off date grouping for now
									return row
								})());
								li.appendChild(currentTimeUl);
								return li
							})());
						}
						currentTimeUl.appendChild(MediaMainLi(media, null, expandedPlaylists));
					}
				}
				ul.replaceChildren(...ulChildren);
				setElChild(main, ul.childNodes.length ? ul : emptyMessageNode);
			}
			else {
				const upperCaseSearchQuery = searchQueryOrNull.toUpperCase();
				ul.replaceChildren(...localMedia.filter((media) => {
					if (!qcsCollabMap.has('media-' + media + '-recentsInfo') || !QcsMediaExists(media)) {
						return false
					}
					if (upperCaseSearchQuery === '') {
						return false
					}
					const subtitle = QcMediaSubtitle(media);
					return (QcMediaTitle(media) + (subtitle === '' ? '' : ' (' + subtitle + ')')).toUpperCase().includes(upperCaseSearchQuery)
				}).slice(0, 50).map((media) => {
					return MediaMainLi(media, null, expandedPlaylists)
				}));
				setElChild(main, ul.childNodes.length ? ul : noSearchResultsMessageNode);
			}
			mountedGlobalPositionsOrNull = globalPositions;
			mountedLocalMediaOrNull = localMedia;
		}
	});
	return main
};
const QcRecentsPage = (tabAfterTick) => {
	const SearchQuery = AnimatedRefFn('');
	const SubmittedSearchQueryOrNull = AnimatedRefFn(null);
	const page = {
		titleBar: (() => {
			const bar = QcTitleBar();
			const input = QcTitleBarInput();
			input.addEventListener('input', () => {
				SearchQuery.set(input.value);
				SubmittedSearchQueryOrNull.set(SearchQuery());
			}, {passive: true});
			animators.set(input, () => {
				input.value = SearchQuery();
			});
			bar.appendChild((() => {
				const form = QcTitleBarForm();
				form.addEventListener('submit', (event) => {
					event.preventDefault();
					input.blur();
					SubmittedSearchQueryOrNull.set(SearchQuery());
				});
				form.appendChild(input);
				return form
			})());
			bar.appendChild((() => {
				const backButton = QcTitleBarButton();
				backButton.addEventListener('click', () => {
					SearchQuery.set('');
					SubmittedSearchQueryOrNull.set(null);
				}, {passive: true});
				backButton.textContent = '⬅️';
				animators.set(backButton, () => {
					backButton.style.display = SearchQuery() === '' && SubmittedSearchQueryOrNull() === null ? 'none' : '';
				});
				return backButton
			})());
			bar.appendChild((() => {
				const clearButton = QcTitleBarButton();
				clearButton.addEventListener('click', () => {
					input.focus();
					SearchQuery.set('');
				}, {passive: true});
				clearButton.textContent = '❌';
				animators.set(clearButton, () => {
					clearButton.style.display = SearchQuery() === '' ? 'none' : '';
				});
				return clearButton
			})());
			return bar
		})(),
		main: (() => {
			const nullMain = Main(null);
			let mainOrNull = null;
			let mainQueryOrNullOrUndefined;
			return () => {
				const queryOrNull = SubmittedSearchQueryOrNull();
				mainOrNull = queryOrNull === null ? nullMain : (mainQueryOrNullOrUndefined === queryOrNull ? mainOrNull : Main(queryOrNull));
				mainQueryOrNullOrUndefined = queryOrNull;
				return mainOrNull
			}
		})(),
	};
	return page
};

const qcRecentsTab = {pageRef: null};
qcRecentsTab.pageRef = new AnimatedRef(QcRecentsPage());

const QcShouldShowYtResult = (result) => {
	if (result.kind === 'channelSection') {
		return result.results.length !== 0
	}
	if (result.kind === 'video') {
		return result.value.duration <= 100000
	}
	return true
};

const AnimatedWeakMap = class {
	constructor () {
		this.weakMap = new WeakMap();
	}
	delete (key) {
		this.weakMap.delete(key);
		animate();
	}
	get (key) {
		return this.weakMap.get(key)
	}
	has (key) {
		return this.weakMap.has(key)
	}
	set (key, value) {
		this.weakMap.set(key, value);
		animate();
	}
};

let time = '00000';
const CollabId = () => {
	const id = collabThread + time;
	time = IncCollabIdInt(time);
	return id
};

const videoResultTracks = new AnimatedWeakMap();
const ExistingTrack = (videoResult) => {
	if (videoResultTracks.has(videoResult)) {
		const track = videoResultTracks.get(videoResult);
		if (QcsMediaExists(track)) {
			return track
		}
	}
	const track = CollabId();
	qcsCollabMap.set('media-' + track + '-title', videoResult.title);
	if (videoResult.subtitleOrNull !== null) {
		qcsCollabMap.set('media-' + track + '-subtitle', videoResult.subtitleOrNull);
	}
	qcsCollabMap.set('media-' + track + '-duration', JSON.stringify(videoResult.value.duration));
	qcsCollabMap.set('media-' + track + '-info2', JSON.stringify({ytVideo: videoResult.value.video}));
	videoResultTracks.set(videoResult, track);
	return track
};
const VideoBumpButton = (result) => {
	const button = QcButton();
	button.addEventListener('click', () => {
		qcBumpTrack(ExistingTrack(result));
	});
	button.appendChild(document.createTextNode('🔼'));
	return button
};
const VideoUnbumpButton = (result) => {
	const button = QcButton();
	button.addEventListener('click', () => {
		if (videoResultTracks.has(result)) {
			qcUnbumpTrack(videoResultTracks.get(result));
		}
	});
	animators.set(button, () => {
		button.style.display = videoResultTracks.has(result) && QcsMediaExists(videoResultTracks.get(result)) && qcsCollabMap.has('media-' + videoResultTracks.get(result) + '-queueInfo') ? '' : 'none';
	});
	button.appendChild(document.createTextNode('🔽'));
	return button
};
const VideoContextButton = (result) => {
	const button = QcMainLiContextButton();
	button.appendChild((() => {
		const queueIndexSpan = document.createElement('span');
		queueIndexSpan.classList.add('mainLiQueueIndexSpan');
		animators.set(queueIndexSpan, () => {
			const indexOrNegativeOne = videoResultTracks.has(result) && qcsCollabMap.has('media-' + videoResultTracks.get(result) + '-queueInfo') ? QcQueuePositions().indexOf(JSON.parse(qcsCollabMap.get('media-' + videoResultTracks.get(result) + '-queueInfo')).position) : -1;
			queueIndexSpan.textContent = indexOrNegativeOne === -1 ? '' : '[' + (QcQueueIsPlaying() ? (indexOrNegativeOne === 0 ? 'Playing' : indexOrNegativeOne.toString(10)) : (indexOrNegativeOne + 1).toString(10)) + '] ';
		});
		return queueIndexSpan
	})());
	button.appendChild(document.createTextNode(result.title));
	if (result.subtitleOrNull !== null) {
		button.appendChild(document.createTextNode(' '));
		button.appendChild((() => {
			const subtitleSpan = document.createElement('span');
			subtitleSpan.classList.add('mainLiSubtitleSpan');
			subtitleSpan.appendChild(document.createTextNode('(' + result.subtitleOrNull + ')'));
			return subtitleSpan
		})());
	}
	return button
};
const PlaylistBumpButton = (result, PlaylistOrNull, videoResultsOrNullRef) => {
	const button = QcButton();
	const LoadOrNull = AnimatedRefFn(null);
	button.addEventListener('click', async () => {
		if (LoadOrNull() === null) {
			const bump = () => {
				if (PlaylistOrNull() === null || !QcsMediaExists(PlaylistOrNull())) {
					const playlist = CollabId();
					qcsCollabMap.set('media-' + playlist + '-title', result.title);
					if (result.subtitleOrNull !== null) {
						qcsCollabMap.set('media-' + playlist + '-subtitle', result.subtitleOrNull);
					}
					let previousPositionOrNull = null;
					for (const videoResult of videoResultsOrNullRef.value) {
						const track = ExistingTrack(videoResult);
						const position = CollabPosition(previousPositionOrNull, null);
						qcsCollabMap.set('media-' + track + '-playlists-' + playlist + '-info', JSON.stringify({position: position}));
						previousPositionOrNull = position;
					}
					qcsCollabMap.set('media-' + playlist + '-info2', JSON.stringify({}));
					PlaylistOrNull.set(playlist);
				}
				qcBumpPlaylist(PlaylistOrNull());
			};
			if (videoResultsOrNullRef.value === null) {
				const load = {};
				LoadOrNull.set(load);
				const videoResultsOrNull = await qcFetchOkJsonOrNull('/ytPlaylistVideoResultsOrNull?' + QueryString({playlist: result.value}));
				if (load === LoadOrNull()) {
					LoadOrNull.set(null);
					if (videoResultsOrNull === null) {
						alert('error message');
					}
					else {
						if (videoResultsOrNullRef.value === null) {
							videoResultsOrNullRef.value = videoResultsOrNull;
						}
						bump();
					}
				}
			}
			else {
				bump();
			}
		}
		else {
			LoadOrNull.set(null);
		}
	});
	animators.set(button, () => {
		button.textContent = LoadOrNull() === null ? '🔼' : '⌛';
	});
	return button
};
const PlaylistUnbumpButton = (result, PlaylistOrNull) => {
	const button = QcButton();
	button.textContent = '🔽';
	button.addEventListener('click', async () => {
		if (PlaylistOrNull() !== null) {
			qcUnbumpPlaylist(PlaylistOrNull());
		}
	});
	animators.set(button, () => {
		const playlistOrNull = PlaylistOrNull();
		button.style.display = playlistOrNull === null ? 'none' : (QcQueuePositions().some((position) => {
			return qcQueueMap.get(position).playlist === playlistOrNull
		}) ? '' : 'none');
	});
	return button
};
const PlaylistContextButton = (result, PlaylistOrNull) => {
	const button = QcMainLiContextButton();
	button.appendChild((() => {
		const queueIndexSpan = document.createElement('span');
		queueIndexSpan.classList.add('mainLiQueueIndexSpan');
		animators.set(queueIndexSpan, () => {
			let text = '';
			if (PlaylistOrNull() !== null) {
				const playlist = PlaylistOrNull();
				const queuePositions = QcQueuePositions();
				const firstIndexOrNegativeOne = queuePositions.findIndex((position) => {
					return qcQueueMap.get(position).playlist === playlist
				});
				if (firstIndexOrNegativeOne !== -1) {
					const firstIndex = firstIndexOrNegativeOne;
					if (firstIndex === 0 && QcQueueIsPlaying()) {
						text = '[Playing] ';
					}
					else {
						let lastIndex;
						for (let i = queuePositions.length - 1; i >= 0; i--) {
							if (qcQueueMap.get(queuePositions[i]).playlist === playlist) {
								lastIndex = i;
								break
							}
						}
						const AdjustedIndex = (index) => {
							return QcQueueIsPlaying() ? index : index + 1
						};
						text = '[' + AdjustedIndex(firstIndex) + (firstIndex === lastIndex ? '' : '-' + AdjustedIndex(lastIndex)) + '] ';
					}
				}
			}
			queueIndexSpan.textContent = text;
		});
		return queueIndexSpan
	})());
	button.appendChild(document.createTextNode(result.title));
	if (result.subtitleOrNull !== null) {
		button.appendChild(document.createTextNode(' '));
		button.appendChild((() => {
			const subtitleSpan = document.createElement('span');
			subtitleSpan.classList.add('mainLiSubtitleSpan');
			subtitleSpan.appendChild(document.createTextNode('(' + result.subtitleOrNull + ')'));
			return subtitleSpan
		})());
	}
	return button
};
const ChannelContextButton = (result) => {
	const button = QcMainLiContextButton();
	button.appendChild(document.createTextNode(result.title));
	if (result.subtitleOrNull !== null) {
		button.appendChild(document.createTextNode(' '));
		button.appendChild((() => {
			const subtitleSpan = document.createElement('span');
			subtitleSpan.classList.add('mainLiSubtitleSpan');
			subtitleSpan.appendChild(document.createTextNode('(' + result.subtitleOrNull + ')'));
			return subtitleSpan
		})());
	}
	return button
};
const QcYtResultLi = (tabAfterTick, pageAfterTick, result) => {
	const li = QcMainLi();
	const isVideoChannelSectionResult = result.kind === 'channelSection' && result.results.every((childResult) => {
		return childResult.kind === 'video'
	});
	const videoResultsOrNullRef = new AnimatedRef(isVideoChannelSectionResult ? result.results : null);
	const isExpandedRef = new AnimatedRef(false);
	li.appendChild((() => {
		const row = document.createElement('div');
		row.classList.add('mainLiRow');
		if (result.kind === 'channel' || (result.kind === 'channelSection' && !isVideoChannelSectionResult) || result.kind === 'channelTab') {
			row.appendChild(ChannelContextButton(result));
			row.appendChild((() => {
				const forwardButton = QcButton();
				const loadOrNullRef = new AnimatedRef(null);
				let childResultsPageOrNull = null;
				queueMicrotask(() => {
					const tab = tabAfterTick;
					const page = pageAfterTick;
					forwardButton.addEventListener('click', async () => {
						if (loadOrNullRef.value === null) {
							if (childResultsPageOrNull === null) {
								const load = {};
								loadOrNullRef.value = load;
								let childResultsOrNull;
								if (result.kind === 'channel') {
									childResultsOrNull = await qcFetchOkJsonOrNull('/ytChannelResultsOrNull3?' + QueryString({channel: result.value, channelTitle: result.title}));
								}
								else if (result.kind === 'channelSection') {
									childResultsOrNull = result.results;
								}
								else if (result.kind === 'channelTab') {
									childResultsOrNull = await qcFetchOkJsonOrNull('/ytChannelTabResultsOrNull?' + QueryString({channel: result.channel, channelTitle: result.subtitleOrNull, tabName: result.tabName}));
								}
								if (load === loadOrNullRef.value) {
									loadOrNullRef.value = null;
									if (childResultsOrNull === null) {
										alert('error message');
									}
									else {
										const childResults = childResultsOrNull;
										const childResultsPage = {
											titleBar: (() => {
												const bar = QcTitleBar();
												bar.appendChild((() => {
													const form = QcTitleBarForm();
													form.appendChild((() => {
														const input = QcTitleBarInput();
														input.disabled = true;
														input.value = result.title + (result.subtitleOrNull === null ? '' : ' (' + result.subtitleOrNull + ')');
														return input
													})());
													return form
												})());
												bar.appendChild((() => {
													const backButton = QcTitleBarButton();
													backButton.addEventListener('click', () => {
														tab.pageRef.value = page;
													}, {passive: true});
													backButton.appendChild(document.createTextNode('⬅️'));
													return backButton
												})());
												return bar
											})(),
											main: null,
										};
										childResultsPage.main = (() => {
											const main = QcMain();
											if (childResults.length === 0) {
												main.appendChild(document.createTextNode((() => {
													if (result.kind === 'channel') {
														return 'empty channel message'
													}
													if (result.kind === 'channelTab') {
														return 'empty channel tab message'
													}
												})()));
											}
											else {
												main.appendChild((() => {
													const ul = QcMainUl();
													for (const childResult of childResults) {
														if (QcShouldShowYtResult(childResult)) {
															ul.appendChild(QcYtResultLi(tab, childResultsPage, childResult));
														}
													}
													return ul
												})());
											}
											return () => {
												return main
											}
										})();
										tab.pageRef.value = childResultsPage;
										childResultsPageOrNull = childResultsPage;
									}
								}
							}
							else {
								tab.pageRef.value = childResultsPageOrNull;
							}
						}
						else {
							loadOrNullRef.value = null;
						}
					}, {passive: true});
				});
				const forwardIcon = document.createTextNode('➡️');
				const loadingIcon = document.createTextNode('⌛');
				animators.set(forwardButton, () => {
					setElChild(forwardButton, loadOrNullRef.value === null ? forwardIcon : loadingIcon);
				});
				return forwardButton
			})());
		}
		else if (result.kind === 'playlist' || isVideoChannelSectionResult) {
			row.appendChild((() => {
				const expandButton = QcButton();
				const loadOrNullRef = new AnimatedRef(null);
				expandButton.addEventListener('click', async () => {
					if (loadOrNullRef.value === null) {
						if (videoResultsOrNullRef.value === null) {
							const load = {};
							loadOrNullRef.value = load;
							const videoResultsOrNull = await qcFetchOkJsonOrNull('/ytPlaylistVideoResultsOrNull?' + QueryString({playlist: result.value}));
							if (load === loadOrNullRef.value) {
								loadOrNullRef.value = null;
								if (videoResultsOrNull === null) {
									alert('error message');
								}
								else {
									if (videoResultsOrNullRef.value === null) {
										videoResultsOrNullRef.value = videoResultsOrNull;
									}
									isExpandedRef.value = true;
								}
							}
						}
						else {
							isExpandedRef.value = !isExpandedRef.value;
						}
					}
					else {
						loadOrNullRef.value = null;
					}
				}, {passive: true});
				const expandIcon = document.createTextNode('➕');
				const loadingIcon = document.createTextNode('⌛');
				const collapseIcon = document.createTextNode('➖');
				animators.set(expandButton, () => {
					setElChild(expandButton, loadOrNullRef.value === null ? (isExpandedRef.value ? collapseIcon : expandIcon) : loadingIcon);
				});
				return expandButton
			})());
			const PlaylistOrNull = AnimatedRefFn(null);
			row.appendChild(PlaylistContextButton(result, PlaylistOrNull));
			row.appendChild(PlaylistUnbumpButton(result, PlaylistOrNull));
			row.appendChild(PlaylistBumpButton(result, PlaylistOrNull, videoResultsOrNullRef));
		}
		else if (result.kind === 'video') {
			row.appendChild(VideoContextButton(result));
			row.appendChild(VideoUnbumpButton(result));
			row.appendChild(VideoBumpButton(result));
		}
		return row
	})());
	if (result.kind === 'playlist' || isVideoChannelSectionResult) {
		li.appendChild((() => {
			const videoResultsUl = document.createElement('ul');
			videoResultsUl.classList.add('mainLiUl');
			let animatedVideoResultsOrNullOrUndefined;
			animators.set(videoResultsUl, () => {
				videoResultsUl.style.display = isExpandedRef.value ? '' : 'none';
				const videoResultsOrNull = videoResultsOrNullRef.value;
				if (animatedVideoResultsOrNullOrUndefined !== videoResultsOrNull) {
					videoResultsUl.replaceChildren();
					if (videoResultsOrNull !== null) {
						for (const videoResult of videoResultsOrNull) {
							if (QcShouldShowYtResult(videoResult)) {
								videoResultsUl.appendChild(QcYtResultLi(tabAfterTick, pageAfterTick, videoResult));
							}
						}
					}
					animatedVideoResultsOrNullOrUndefined = videoResultsOrNull;
				}
			});
			return videoResultsUl
		})());
	}
	return li
};

const WelcomeMain = () => {
	const main = QcMain();
	main.appendChild(document.createTextNode('yt search page message'));
	return main
};
const ResultsMain = (tabAfterTick, pageAfterTick, results) => {
	const main = QcMain();
	main.appendChild(results.length === 0 ? document.createTextNode('no results message') : (() => {
		const ul = QcMainUl();
		for (const result of results) {
			if (QcShouldShowYtResult(result)) {
				ul.appendChild(QcYtResultLi(tabAfterTick, pageAfterTick, result));
			}
		}
		return ul
	})());
	return main
};
const QcYtSearchPage = (tabAfterTick) => {
	const mainRef = new AnimatedRef(WelcomeMain());
	const page = {
		titleBar: (() => {
			const bar = QcTitleBar();
			const input = QcTitleBarInput();
			const queryRef = new AnimatedRef('');
			input.addEventListener('input', () => {
				queryRef.value = input.value;
			}, {passive: true});
			animators.set(input, () => {
				input.value = queryRef.value;
			});
			const loadOrNullRef = new AnimatedRef(null);
			bar.appendChild((() => {
				const form = QcTitleBarForm();
				form.addEventListener('submit', async (event) => {
					event.preventDefault();
					input.blur();
					if (queryRef.value === '') {
						loadOrNullRef.value = null;
						mainRef.value = WelcomeMain();
					}
					else {
						const load = {};
						loadOrNullRef.value = load;
						const resultsOrNull = await qcFetchOkJsonOrNull('/ytSearchResultsOrNull?' + QueryString({query: queryRef.value}));
						if (load === loadOrNullRef.value) {
							loadOrNullRef.value = null;
							if (resultsOrNull === null) {
								alert('error message');
							}
							else {
								mainRef.value = ResultsMain(tabAfterTick, page, resultsOrNull);
							}
						}
					}
				});
				form.appendChild(input);
				return form
			})());
			bar.appendChild((() => {
				const button = QcTitleBarButton();
				button.addEventListener('click', () => {
					if (loadOrNullRef.value === null) {
						input.focus();
						queryRef.value = '';
					}
					else {
						loadOrNullRef.value = null;
					}
				}, {passive: true});
				const clearIcon = document.createTextNode('❌');
				const loadingIcon = document.createTextNode('⌛');
				animators.set(button, () => {
					setElChild(button, loadOrNullRef.value === null ? clearIcon : loadingIcon);
					button.style.display = loadOrNullRef.value === null ? (queryRef.value === '' ? 'none' : '') : '';
				});
				return button
			})());
			return bar
		})(),
		main: () => {
			return mainRef.value
		},
	};
	return page
};

const qcYtTab = {pageRef: null};
qcYtTab.pageRef = new AnimatedRef(QcYtSearchPage(qcYtTab));

document.body.classList.add('body');
let animationMain;
const mainScrollInfo = new WeakMap();
animators.set(document.body, () => {
	animationMain = qcCurrentTabRef.value.pageRef.value.main();
	const mountedMainOrNull = QcMountedMainOrNull();
	if (mountedMainOrNull !== animationMain && mountedMainOrNull !== null) {
		mainScrollInfo.set(mountedMainOrNull, {left: document.documentElement.scrollLeft, top: document.documentElement.scrollTop});
	}
});
document.body.appendChild((() => {
	const header = document.createElement('header');
	header.appendChild((() => {
		const topBar = document.createElement('div');
		topBar.classList.add('topBar');
		topBar.appendChild((() => {
			const tabBar = document.createElement('nav');
			tabBar.classList.add('tabBar');
			const Button = (tab, name) => {
				const button = QcButton();
				button.classList.add('tabButton');
				button.addEventListener('click', () => {
					qcCurrentTabRef.value = tab;
				}, {passive: true});
				button.appendChild(document.createTextNode(name));
				animators.set(button, () => {
					button.classList.toggle('currentTabButton', tab === qcCurrentTabRef.value);
				});
				return button
			};
			tabBar.appendChild(Button(qcQueueTab, 'Queue'));
			tabBar.appendChild(Button(qcRecentsTab, 'Recents'));
			tabBar.appendChild(Button(qcYtTab, 'YouTube'));
			return tabBar
		})());
		topBar.appendChild((() => {
			const titleBarDiv = document.createElement('div');
			animators.set(titleBarDiv, () => {
				setElChild(titleBarDiv, qcCurrentTabRef.value.pageRef.value.titleBar);
			});
			return titleBarDiv
		})());
		return topBar
	})());
	header.appendChild((() => {
		const logoButtonDiv = document.createElement('div');
		logoButtonDiv.classList.add('logoButtonDiv');
		logoButtonDiv.appendChild((() => {
			const logoButton = QcButton();
			logoButton.classList.add('logoButton');
			logoButton.addEventListener('click', async () => {
				const uriOrNull = await qcFetchOkJsonOrNull('/remoteQcUriOrNull');
				if (uriOrNull === null) {
					alert('Something went wrong getting the link :(');
				}
				else {
					alert(uriOrNull);
					// if (confirm('The link to this queue is '+uriOrNull+'. Press OK to copy it to the clipboard.')) {
					// 	console.log('copy')
					// }else {
					// 	console.log('nocopu')
					// }
				}
			});
			logoButton.appendChild(document.createTextNode('🔗'));
			return logoButton
		})());
		return logoButtonDiv
	})());
	return header
})());
document.body.appendChild((() => {
	const mainDiv = document.createElement('div');
	animators.set(mainDiv, () => {
		setElChild(mainDiv, animationMain);
	});
	return mainDiv
})());
postAnimators.set(document.body, () => {
	if (QcMountedMainOrNull() !== animationMain) {
		if (mainScrollInfo.has(animationMain)) {
			const info = mainScrollInfo.get(animationMain);
			document.documentElement.scroll(info.left, info.top);
		}
		else {
			document.documentElement.scroll(0, 0);
		}
		QcMountedMainOrNull.set(animationMain);
	}
});
animate();

const Queue = class {
	constructor () {
		this.firstNodeOrNull = null;
		this.lastNodeOrNull = null;
	}
	append (value) {
		const node = {nextNodeOrNull: null, value: value};
		if (this.lastNodeOrNull === null) {
			this.firstNodeOrNull = node;
		}
		else {
			this.lastNodeOrNull.nextNodeOrNull = node;
		}
		this.lastNodeOrNull = node;
	}
	deleteFirstValue () {
		this.firstNodeOrNull = this.firstNodeOrNull.nextNodeOrNull;
		if (this.firstNodeOrNull === null) {
			this.lastNodeOrNull = null;
		}
	}
	firstValue () {
		return this.firstNodeOrNull.value
	}
	isEmpty () {
		return this.firstNodeOrNull === null
	}
};

const wait = (delay) => {
	return new Promise((resolve) => {
		setTimeout(resolve, delay);
	})
};

let qsCollabMapTime = qcSetupArg.qsCollabMapTime;
const qsCollabThread = qcSetupArg.qsCollabThread;
let currentWs;
let currentWsUnconfirmedCollabMapOps;
const collabMapOpsToSync = new Map()
;(async () => {
	while (true) {
		const ws = new WebSocket('ws://' + location.host + '/qcWs?' + QueryString({
			qsCollabMapTime: JSON.stringify(qsCollabMapTime),
			qsCollabThread: qsCollabThread,
		}));
		currentWs = ws;
		const wsUnconfirmedCollabMapOps = new Queue();
		currentWsUnconfirmedCollabMapOps = wsUnconfirmedCollabMapOps;
		let openTimeout;
		await new Promise((resolve) => {
			openTimeout = setTimeout(() => {
				if (ws.readyState === 0) {
					ws.close();
					resolve();
				}
			}, 10000);
			let wsOpened = false;
			ws.addEventListener('open', () => {
				wsOpened = true;
				for (const op of collabMapOpsToSync.values()) {
					ws.send(JSON.stringify({kind: 'collabMapOp', value: op}));
					wsUnconfirmedCollabMapOps.append(op);
				}
				let wsHasReceivedMessageSinceLastPing = false
				;(async () => {
					while (true) {
						ws.send(JSON.stringify({kind: 'ping', value: null}));
						await wait(1000); // todo fix this timeout leak
						if (!wsHasReceivedMessageSinceLastPing) {
							ws.close();
							resolve();
							break
						}
						wsHasReceivedMessageSinceLastPing = false;
					}
				})();
				ws.addEventListener('message', (event) => {
					wsHasReceivedMessageSinceLastPing = true;
					const message = JSON.parse(event.data);
					const kind = message.kind;
					const value = message.value;
					if (kind === 'collabMapCatchUpOp') {
						qcsCollabMap.receive(value);
					}
					else if (kind === 'collabMapCatchUpTime') {
						qsCollabMapTime = value;
					}
					else if (kind === 'collabMapOp') {
						qcsCollabMap.receive(value);
						qsCollabMapTime++;
					}
					else if (kind === 'collabMapOpConfirmation') {
						const op = wsUnconfirmedCollabMapOps.firstValue();
						wsUnconfirmedCollabMapOps.deleteFirstValue();
						if (collabMapOpsToSync.get(op.key) === op) {
							collabMapOpsToSync.delete(op.key);
						}
					}
					else if (kind === 'qsCollabThreadMismatch') {
						location.reload();
					}
				}, {passive: true});
			}, {passive: true});
			ws.addEventListener('close', () => {
				if (wsOpened) {
					resolve();
				}
				else {
					setTimeout(resolve, 1000);
				}
			}, {passive: true});
		});
		clearTimeout(openTimeout);
	}
})();
const handleNewCollabMapOp = (op) => {
	if (qcsCollabMap.localOps.has(op)) {
		collabMapOpsToSync.set(op.key, op);
		if (currentWs.readyState === 1) {
			currentWs.send(JSON.stringify({kind: 'collabMapOp', value: op}));
			currentWsUnconfirmedCollabMapOps.append(op);
		}
	}
};
for (const op of qcsCollabMap.ops.values()) {
	handleNewCollabMapOp(op);
}
qcsCollabMap.changeHandlers.push((change) => {
	handleNewCollabMapOp(change.op);
});

window.addEventListener('error', async (event) => {
	try {
		// todo ErrorEventAsJsonValue, see https://developer.mozilla.org/en-US/docs/Web/API/ErrorEvent
		await qcWriteDebuggingInfo('qcError', ThroweeAsJsonValue(event.error));
	}
	catch (throwee) {
	}
}, {passive: true});
window.addEventListener('unhandledrejection', async (event) => {
	try {
		await qcWriteDebuggingInfo('qcPromiseRejectionReason', ThroweeAsJsonValue(event.reason));
	}
	catch (throwee) {
	}
}, {passive: true});
qcWriteDebuggingInfo('qcInfo', {
	date: DateAsJsonValue(new Date()),
	locationHref: location.href,
	navigatorUserAgent: navigator.userAgent,
	qcSetupArgQsCollabThread: qcSetupArg.qsCollabThread,
	isSecureContext: isSecureContext,
});
if (location.hostname === 'localhost') {
	const audio = new Audio();
	audio.addEventListener('play', () => {
		if (!QcQueueIsPlaying()) {
			QcQueueIsPlaying.set(true);
		}
	});
	audio.addEventListener('pause', () => {
		if (QcQueueIsPlaying() && !audio.ended) { // in some browsers paused is set to true and pause event fires when audio ends ¯\_(ツ)_/¯
			QcQueueIsPlaying.set(false);
		}
	});
	audio.addEventListener('ended', () => {
		qcSkipForward();
	});
	let currentTrackOrNull = null;
	let audioUpdateIsQueued = false;
	const queueAudioUpdate = () => {
		if (!audioUpdateIsQueued) {
			queueMicrotask(() => {
				audioUpdateIsQueued = false;
				const positions = QcQueuePositions();
				const newTrackOrNull = positions.length === 0 ? null : qcQueueMap.get(positions[0]).track;
				if (currentTrackOrNull !== newTrackOrNull) {
					audio.src = newTrackOrNull === null ? '' : '/ytTrackAudio2?' + QueryString({track: newTrackOrNull, video: QcsMediaInfoOrNull(newTrackOrNull).ytVideo});
					currentTrackOrNull = newTrackOrNull;
				}
				if (QcQueueIsPlaying()) {
					if (audio.paused) {
						audio.play();
					}
				}
				else {
					if (!audio.paused) {
						audio.pause();
					}
				}
				audio.volume = QcQueueVolume();
			});
			audioUpdateIsQueued = true;
		}
	};
	queueAudioUpdate();
	qcsCollabMap.changeHandlers.push(queueAudioUpdate);
}
